#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""PyQt5 module to provide multithreaded communication and periodical data
acquisition for a Thermo Scientific ThermoFlex recirculating chiller.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "07-07-2020"  # 0.0.1 was stamped 18-09-2018
__version__ = "0.0.5"  # 0.0.1 corresponds to prototype 1.0.0

from PyQt5 import QtCore, QtGui
from PyQt5 import QtWidgets as QtWid

from dvg_utils.dvg_pyqt_controls import (
    create_Toggle_button,
    create_error_LED,
    create_tiny_error_LED,
    SS_GROUP,
)
from dvg_debug_functions import print_fancy_traceback as pft

from dvg_qdeviceio import QDeviceIO, DAQ_TRIGGER
from dvg_devices.ThermoFlex_chiller_protocol_RS232 import ThermoFlex_chiller

# Special characters
CHAR_DEG_C = chr(176) + "C"


class ThermoFlex_chiller_qdev(QDeviceIO):
    """Manages multithreaded communication and periodical data acquisition for
    a Thermo Scientific ThermoFlex recirculating chiller, referred to as the
    'device'.

    In addition, it also provides PyQt5 GUI objects for control of the device.
    These can be incorporated into your application.

    All device I/O operations will be offloaded to 'workers', each running in
    a newly created thread instead of in the main/GUI thread.

        - Worker_DAQ:
            Periodically acquires data from the device.

        - Worker_jobs:
            Maintains a thread-safe queue where desired device I/O operations
            can be put onto, and sends the queued operations first in first out
            (FIFO) to the device.

    (*): See 'dvg_qdeviceio.QDeviceIO()' for details.

    Args:
        dev:
            Reference to a
            'dvg_devices.ThermoFlex_chiller_protocol_RS232.ThermoFlex_chiller'
            instance.

        (*) DAQ_interval_ms
        (*) critical_not_alive_count
        (*) DAQ_timer_type

    Main methods:
        (*) start(...)
        (*) quit()

    Main data attributes:
        (*) DAQ_update_counter
        (*) obtained_DAQ_interval_ms
        (*) obtained_DAQ_rate_Hz

    Main GUI objects:
        hbly_GUI (PyQt5.QtWidgets.QHBoxLayout)

    Signals:
        (*) signal_DAQ_updated()
        (*) signal_connection_lost()
    """

    signal_GUI_alarm_values_update = QtCore.pyqtSignal()
    signal_GUI_PID_values_update = QtCore.pyqtSignal()

    def __init__(
        self,
        dev: ThermoFlex_chiller,
        DAQ_interval_ms=1000,
        DAQ_timer_type=QtCore.Qt.CoarseTimer,
        critical_not_alive_count=1,
        calc_DAQ_rate_every_N_iter=1,
        debug=False,
        **kwargs,
    ):
        super().__init__(dev, **kwargs)  # Pass kwargs onto QtCore.QObject()

        self.create_worker_DAQ(
            DAQ_trigger=DAQ_TRIGGER.INTERNAL_TIMER,
            DAQ_function=self.DAQ_function,
            DAQ_interval_ms=DAQ_interval_ms,
            DAQ_timer_type=DAQ_timer_type,
            critical_not_alive_count=critical_not_alive_count,
            calc_DAQ_rate_every_N_iter=calc_DAQ_rate_every_N_iter,
            debug=debug,
        )

        self.create_worker_jobs(jobs_function=self.jobs_function, debug=debug)

        self.create_GUI()
        self.signal_DAQ_updated.connect(self.update_GUI)
        self.connect_signals_to_slots()
        if not self.dev.is_alive:
            self.update_GUI()  # Correctly reflect an offline device

    # --------------------------------------------------------------------------
    #   DAQ_function
    # --------------------------------------------------------------------------

    def DAQ_function(self):
        success = self.dev.query_status_bits()
        success &= self.dev.query_state()

        return success

    # --------------------------------------------------------------------------
    #   jobs_function
    # --------------------------------------------------------------------------

    def jobs_function(self, func, args):
        if func == "signal_GUI_alarm_values_update":
            # Special instruction
            self.signal_GUI_alarm_values_update.emit()
        elif func == "signal_GUI_PID_values_update":
            # Special instruction
            self.signal_GUI_PID_values_update.emit()
        else:
            # Default job processing:
            # Send I/O operation to the device
            try:
                func(*args)
            except Exception as err:
                pft(err)

    # --------------------------------------------------------------------------
    #   create_GUI
    # --------------------------------------------------------------------------

    def create_GUI(self):
        # Groupbox "Alarm values"
        # -----------------------
        p = {
            "alignment": QtCore.Qt.AlignRight,
            "minimumWidth": 50,
            "maximumWidth": 30,
            "readOnly": True,
        }
        self.LO_flow = QtWid.QLineEdit(**p)
        self.HI_flow = QtWid.QLineEdit(**p)
        self.LO_pres = QtWid.QLineEdit(**p)
        self.HI_pres = QtWid.QLineEdit(**p)
        self.LO_temp = QtWid.QLineEdit(**p)
        self.HI_temp = QtWid.QLineEdit(**p)
        self.pbtn_read_alarm_values = QtWid.QPushButton("Read")
        self.pbtn_read_alarm_values.setMinimumSize(50, 30)

        p = {"alignment": QtCore.Qt.AlignCenter}
        grid = QtWid.QGridLayout()
        # fmt: off
        grid.addWidget(QtWid.QLabel("Values can be set in the chiller's menu",
                                    **p)          , 0, 0, 1, 4)
        grid.addWidget(QtWid.QLabel("LO")         , 1, 1)
        grid.addWidget(QtWid.QLabel("HI")         , 1, 2)
        grid.addWidget(QtWid.QLabel("Flow rate")  , 2, 0)
        grid.addWidget(self.LO_flow               , 2, 1)
        grid.addWidget(self.HI_flow               , 2, 2)
        grid.addWidget(QtWid.QLabel("LPM")        , 2, 3)
        grid.addWidget(QtWid.QLabel("Pressure")   , 3, 0)
        grid.addWidget(self.LO_pres               , 3, 1)
        grid.addWidget(self.HI_pres               , 3, 2)
        grid.addWidget(QtWid.QLabel("bar")        , 3, 3)
        grid.addWidget(QtWid.QLabel("Temperature"), 4, 0)
        grid.addWidget(self.LO_temp               , 4, 1)
        grid.addWidget(self.HI_temp               , 4, 2)
        grid.addWidget(QtWid.QLabel(CHAR_DEG_C)   , 4, 3)
        grid.addWidget(self.pbtn_read_alarm_values, 5, 0)
        # fmt: on

        self.grpb_alarms = QtWid.QGroupBox("Alarm values")
        self.grpb_alarms.setStyleSheet(SS_GROUP)
        self.grpb_alarms.setLayout(grid)

        # Groupbox "PID feedback"
        # -----------------------
        p = {
            "alignment": QtCore.Qt.AlignRight,
            "minimumWidth": 50,
            "maximumWidth": 30,
            "readOnly": True,
        }
        self.PID_P = QtWid.QLineEdit(**p)
        self.PID_I = QtWid.QLineEdit(**p)
        self.PID_D = QtWid.QLineEdit(**p)
        self.pbtn_read_PID_values = QtWid.QPushButton("Read")
        self.pbtn_read_PID_values.setMinimumSize(50, 30)

        p = {"alignment": QtCore.Qt.AlignCenter}
        grid = QtWid.QGridLayout()
        # fmt: off
        grid.addWidget(QtWid.QLabel("Values can be set in the chiller's menu",
                                    **p)                          , 0, 0, 1, 3)
        grid.addWidget(QtWid.QLabel("P", **p)                     , 1, 0)
        grid.addWidget(self.PID_P                                 , 1, 1)
        grid.addWidget(QtWid.QLabel("% span of 100" + CHAR_DEG_C) , 1, 2)
        grid.addWidget(QtWid.QLabel("I", **p)                     , 2, 0)
        grid.addWidget(self.PID_I                                 , 2, 1)
        grid.addWidget(QtWid.QLabel("repeats/minute")             , 2, 2)
        grid.addWidget(QtWid.QLabel("D", **p)                     , 3, 0)
        grid.addWidget(self.PID_D                                 , 3, 1)
        grid.addWidget(QtWid.QLabel("minutes")                    , 3, 2)
        grid.addWidget(self.pbtn_read_PID_values                  , 4, 0)
        # fmt: on

        self.grpb_PID = QtWid.QGroupBox("PID feedback")
        self.grpb_PID.setStyleSheet(SS_GROUP)
        self.grpb_PID.setLayout(grid)

        # Groupbox "Status bits"
        # ----------------------
        # fmt: off
        self.SB_tripped                = create_error_LED()
        self.SB_tripped.setText("No faults")
        self.SB_high_temp_fixed        = create_tiny_error_LED()
        self.SB_low_temp_fixed         = create_tiny_error_LED()
        self.SB_high_temp              = create_tiny_error_LED()
        self.SB_low_temp               = create_tiny_error_LED()
        self.SB_high_pressure          = create_tiny_error_LED()
        self.SB_low_pressure           = create_tiny_error_LED()
        self.SB_drip_pan               = create_tiny_error_LED()
        self.SB_high_level             = create_tiny_error_LED()
        self.SB_phase_monitor          = create_tiny_error_LED()
        self.SB_motor_overload         = create_tiny_error_LED()
        self.SB_LPC                    = create_tiny_error_LED()
        self.SB_HPC                    = create_tiny_error_LED()
        self.SB_external_EMO           = create_tiny_error_LED()
        self.SB_local_EMO              = create_tiny_error_LED()
        self.SB_low_flow               = create_tiny_error_LED()
        self.SB_low_level              = create_tiny_error_LED()
        self.SB_sense_5V               = create_tiny_error_LED()
        self.SB_invalid_level          = create_tiny_error_LED()
        self.SB_low_fixed_flow_warning = create_tiny_error_LED()
        self.SB_high_pressure_factory  = create_tiny_error_LED()
        self.SB_low_pressure_factory   = create_tiny_error_LED()

        p = {'alignment': QtCore.Qt.AlignRight}
        grid = QtWid.QGridLayout()
        grid.addWidget(self.SB_tripped                            , 0, 0, 1, 2)
        grid.addItem(QtWid.QSpacerItem(1, 12)                          , 1, 0)
        grid.addWidget(QtWid.QLabel("high temp fixed fault", **p)      , 2, 0)
        grid.addWidget(self.SB_high_temp_fixed                         , 2, 1)
        grid.addWidget(QtWid.QLabel("low temp fixed fault", **p)       , 3, 0)
        grid.addWidget(self.SB_low_temp_fixed                          , 3, 1)
        grid.addWidget(QtWid.QLabel("high temp fault/warning", **p)    , 4, 0)
        grid.addWidget(self.SB_high_temp                               , 4, 1)
        grid.addWidget(QtWid.QLabel("low temp fault/warning", **p)     , 5, 0)
        grid.addWidget(self.SB_low_temp                                , 5, 1)
        grid.addWidget(QtWid.QLabel("high pressure fault/warning", **p), 6, 0)
        grid.addWidget(self.SB_high_pressure                           , 6, 1)
        grid.addWidget(QtWid.QLabel("low pressure fault/warning", **p) , 7, 0)
        grid.addWidget(self.SB_low_pressure                            , 7, 1)
        grid.addWidget(QtWid.QLabel("drip pan fault", **p)             , 8, 0)
        grid.addWidget(self.SB_drip_pan                                , 8, 1)
        grid.addWidget(QtWid.QLabel("high level fault", **p)           , 9, 0)
        grid.addWidget(self.SB_high_level                              , 9, 1)
        grid.addWidget(QtWid.QLabel("phase monitor fault", **p)        , 10, 0)
        grid.addWidget(self.SB_phase_monitor                           , 10, 1)
        grid.addWidget(QtWid.QLabel("motor overload fault", **p)       , 11, 0)
        grid.addWidget(self.SB_motor_overload                          , 11, 1)
        grid.addWidget(QtWid.QLabel("LPC fault", **p)                  , 12, 0)
        grid.addWidget(self.SB_LPC                                     , 12, 1)
        grid.addWidget(QtWid.QLabel("HPC fault", **p)                  , 13, 0)
        grid.addWidget(self.SB_HPC                                     , 13, 1)
        grid.addWidget(QtWid.QLabel("external EMO fault", **p)         , 14, 0)
        grid.addWidget(self.SB_external_EMO                            , 14, 1)
        grid.addWidget(QtWid.QLabel("local EMO fault", **p)            , 15, 0)
        grid.addWidget(self.SB_local_EMO                               , 15, 1)
        grid.addWidget(QtWid.QLabel("low flow fault", **p)             , 16, 0)
        grid.addWidget(self.SB_low_flow                                , 16, 1)
        grid.addWidget(QtWid.QLabel("low level fault", **p)            , 17, 0)
        grid.addWidget(self.SB_low_level                               , 17, 1)
        grid.addWidget(QtWid.QLabel("sense 5V fault", **p)             , 18, 0)
        grid.addWidget(self.SB_sense_5V                                , 18, 1)
        grid.addWidget(QtWid.QLabel("invalid level fault", **p)        , 19, 0)
        grid.addWidget(self.SB_invalid_level                           , 19, 1)
        grid.addWidget(QtWid.QLabel("low fixed flow warning", **p)     , 20, 0)
        grid.addWidget(self.SB_low_fixed_flow_warning                  , 20, 1)
        grid.addWidget(QtWid.QLabel("high pressure factory fault", **p), 21, 0)
        grid.addWidget(self.SB_high_pressure_factory                   , 21, 1)
        grid.addWidget(QtWid.QLabel("low pressure factory fault", **p) , 22, 0)
        grid.addWidget(self.SB_low_pressure_factory                    , 22, 1)
        # fmt: on

        self.grpb_SBs = QtWid.QGroupBox("Status bits")
        self.grpb_SBs.setStyleSheet(SS_GROUP)
        self.grpb_SBs.setLayout(grid)

        # Groupbox "Control"
        # ------------------
        p = {
            "alignment": QtCore.Qt.AlignRight,
            "minimumWidth": 50,
            "maximumWidth": 30,
        }

        self.lbl_offline = QtWid.QLabel(
            "OFFLINE",
            visible=False,
            font=QtGui.QFont("Palatino", 14, weight=QtGui.QFont.Bold),
            alignment=QtCore.Qt.AlignCenter,
        )
        # fmt: off
        self.pbtn_on       = create_Toggle_button("Off")
        self.powering_down = create_tiny_error_LED()
        self.send_setpoint = QtWid.QLineEdit(**p)
        self.read_setpoint = QtWid.QLineEdit(**p, readOnly=True)
        self.read_temp     = QtWid.QLineEdit(**p, readOnly=True)
        self.read_flow     = QtWid.QLineEdit(**p, readOnly=True)
        self.read_supply   = QtWid.QLineEdit(**p, readOnly=True)
        self.read_suction  = QtWid.QLineEdit(**p, readOnly=True)
        self.lbl_update_counter = QtWid.QLabel("0")

        grid = QtWid.QGridLayout()
        grid.addWidget(self.lbl_offline                   , 0, 0, 1, 3)
        grid.addWidget(self.pbtn_on                       , 1, 0, 1, 3)
        grid.addWidget(QtWid.QLabel("Is powering up/down?",
                       alignment=QtCore.Qt.AlignRight)    , 2, 0, 1, 2)
        grid.addWidget(self.powering_down                 , 2, 2)
        grid.addItem(QtWid.QSpacerItem(1, 12)             , 3, 0)
        grid.addWidget(QtWid.QLabel("Send setpoint")      , 4, 0)
        grid.addWidget(QtWid.QLabel("Read setpoint")      , 5, 0)
        grid.addWidget(self.send_setpoint                 , 4, 1)
        grid.addWidget(self.read_setpoint                 , 5, 1)
        grid.addWidget(QtWid.QLabel(CHAR_DEG_C)           , 4, 2)
        grid.addWidget(QtWid.QLabel(CHAR_DEG_C)           , 5, 2)
        grid.addItem(QtWid.QSpacerItem(1, 12)             , 6, 0)
        grid.addWidget(QtWid.QLabel("Read temp")          , 7, 0)
        grid.addWidget(self.read_temp                     , 7, 1)
        grid.addWidget(QtWid.QLabel(CHAR_DEG_C)           , 7, 2)
        grid.addWidget(QtWid.QLabel("Read flow")          , 8, 0)
        grid.addWidget(self.read_flow                     , 8, 1)
        grid.addWidget(QtWid.QLabel("LPM")                , 8, 2)
        grid.addWidget(QtWid.QLabel("Read supply")        , 9, 0)
        grid.addWidget(self.read_supply                   , 9, 1)
        grid.addWidget(QtWid.QLabel("bar")                , 9, 2)
        grid.addWidget(QtWid.QLabel("Read suction")       , 10, 0)
        grid.addWidget(self.read_suction                  , 10, 1)
        grid.addWidget(QtWid.QLabel("bar")                , 10, 2)

        grid.addItem(QtWid.QSpacerItem(1, 12)                      , 11, 0)
        grid.addWidget(QtWid.QLabel("Nominal values @ 15-02-2018:"), 12, 0, 1, 3)
        grid.addWidget(QtWid.QLabel("Read flow")                   , 13, 0)
        grid.addWidget(QtWid.QLabel("80  ",
                                    alignment=QtCore.Qt.AlignRight), 13, 1)
        grid.addWidget(QtWid.QLabel("LPM")                         , 13, 2)
        grid.addWidget(QtWid.QLabel("Read supply")                 , 14, 0)
        grid.addWidget(QtWid.QLabel("2.9  ",
                                    alignment=QtCore.Qt.AlignRight), 14, 1)
        grid.addWidget(QtWid.QLabel("bar")                         , 14, 2)
        grid.addWidget(QtWid.QLabel("Read suction")                , 15, 0)
        grid.addWidget(QtWid.QLabel("40  ",
                                    alignment=QtCore.Qt.AlignRight), 15, 1)
        grid.addWidget(QtWid.QLabel("bar")                         , 15, 2)
        grid.addWidget(self.lbl_update_counter                     , 16, 0, 1, 2)
        # fmt: on

        self.grpb_control = QtWid.QGroupBox("Control")
        self.grpb_control.setStyleSheet(SS_GROUP)
        self.grpb_control.setLayout(grid)

        # --------------------------------------
        #   Round up final QtWid.QHBoxLayout()
        # --------------------------------------

        vbox = QtWid.QVBoxLayout()
        vbox.addWidget(self.grpb_alarms)
        vbox.addWidget(self.grpb_PID)
        vbox.setAlignment(self.grpb_alarms, QtCore.Qt.AlignTop)
        vbox.setAlignment(self.grpb_PID, QtCore.Qt.AlignTop)
        vbox.setAlignment(QtCore.Qt.AlignTop)

        self.hbly_GUI = QtWid.QHBoxLayout()
        self.hbly_GUI.addLayout(vbox)
        self.hbly_GUI.addWidget(self.grpb_SBs)
        self.hbly_GUI.addWidget(self.grpb_control)
        self.hbly_GUI.addStretch(1)
        self.hbly_GUI.setAlignment(self.grpb_SBs, QtCore.Qt.AlignTop)
        self.hbly_GUI.setAlignment(self.grpb_control, QtCore.Qt.AlignTop)
        self.hbly_GUI.setAlignment(QtCore.Qt.AlignTop)

        # tab_chiller.setLayout(self.hbly_GUI)

    # --------------------------------------------------------------------------
    #   update_GUI
    # --------------------------------------------------------------------------

    @QtCore.pyqtSlot()
    def update_GUI(self):
        """NOTE: 'self.dev.mutex' is not being locked, because we are only
        reading 'state' for displaying purposes. We can do this because 'state'
        members are written and read atomicly.
        Not locking the mutex might speed up the program.
        """
        if self.dev.is_alive:
            # At startup
            if self.DAQ_update_counter == 1:
                self.update_GUI_alarm_values()
                self.update_GUI_PID_values()
                self.send_setpoint.setText("%.1f" % self.dev.state.setpoint)

            # State
            # fmt: off
            self.read_setpoint.setText("%.1f" % self.dev.state.setpoint)
            self.read_temp.setText    ("%.1f" % self.dev.state.temp)
            self.read_flow.setText    ("%.1f" % self.dev.state.flow)
            self.read_supply.setText  ("%.2f" % self.dev.state.supply_pres)
            self.read_suction.setText ("%.2f" % self.dev.state.suction_pres)
            # fmt: on

            # Power
            SBs = self.dev.status_bits  # Short-hand
            self.pbtn_on.setChecked(SBs.running)
            if SBs.running:
                self.pbtn_on.setText("ON")
            else:
                self.pbtn_on.setText("OFF")
            self.powering_down.setChecked(SBs.powering_down)

            # Status bits
            self.SB_tripped.setChecked(SBs.fault_tripped)
            if self.dev.status_bits.fault_tripped:
                self.SB_tripped.setText("FAULT TRIPPED")
            else:
                self.SB_tripped.setText("No faults")
            self.SB_drip_pan.setChecked(SBs.drip_pan_fault)
            self.SB_external_EMO.setChecked(SBs.external_EMO_fault)
            self.SB_high_level.setChecked(SBs.high_level_fault)
            self.SB_high_pressure.setChecked(SBs.high_pressure_fault)
            self.SB_high_pressure_factory.setChecked(
                SBs.high_pressure_fault_factory
            )
            self.SB_high_temp.setChecked(SBs.high_temp_fault)
            self.SB_high_temp_fixed.setChecked(SBs.high_temp_fixed_fault)
            self.SB_HPC.setChecked(SBs.HPC_fault)
            self.SB_invalid_level.setChecked(SBs.invalid_level_fault)
            self.SB_local_EMO.setChecked(SBs.local_EMO_fault)
            self.SB_low_fixed_flow_warning.setChecked(
                SBs.low_fixed_flow_warning
            )
            self.SB_low_flow.setChecked(SBs.low_flow_fault)
            self.SB_low_level.setChecked(SBs.low_level_fault)
            self.SB_low_pressure.setChecked(SBs.low_pressure_fault)
            self.SB_low_pressure_factory.setChecked(
                SBs.low_pressure_fault_factory
            )
            self.SB_low_temp.setChecked(SBs.low_temp_fault)
            self.SB_low_temp_fixed.setChecked(SBs.low_temp_fixed_fault)
            self.SB_LPC.setChecked(SBs.LPC_fault)
            self.SB_motor_overload.setChecked(SBs.motor_overload_fault)
            self.SB_phase_monitor.setChecked(SBs.phase_monitor_fault)
            self.SB_sense_5V.setChecked(SBs.sense_5V_fault)

            self.lbl_update_counter.setText("%s" % self.DAQ_update_counter)
        else:
            self.grpb_alarms.setEnabled(False)
            self.grpb_PID.setEnabled(False)
            self.grpb_SBs.setEnabled(False)
            self.grpb_control.setEnabled(False)

            self.pbtn_on.setVisible(False)
            self.lbl_offline.setVisible(True)

    @QtCore.pyqtSlot()
    def update_GUI_alarm_values(self):
        self.LO_flow.setText("%.1f" % self.dev.values_alarm.LO_flow)
        if self.dev.values_alarm.HI_flow == 0:
            self.HI_flow.setText("No limit")
        else:
            self.HI_flow.setText("%.1f" % self.dev.values_alarm.HI_flow)
        self.LO_pres.setText("%.2f" % self.dev.values_alarm.LO_pres)
        self.HI_pres.setText("%.2f" % self.dev.values_alarm.HI_pres)
        self.LO_temp.setText("%.1f" % self.dev.values_alarm.LO_temp)
        self.HI_temp.setText("%.1f" % self.dev.values_alarm.HI_temp)

    @QtCore.pyqtSlot()
    def update_GUI_PID_values(self):
        self.PID_P.setText("%.1f" % self.dev.values_PID.P)
        self.PID_I.setText("%.2f" % self.dev.values_PID.I)
        self.PID_D.setText("%.1f" % self.dev.values_PID.D)

    # --------------------------------------------------------------------------
    #   GUI functions
    # --------------------------------------------------------------------------

    @QtCore.pyqtSlot()
    def process_pbtn_on(self):
        if self.dev.status_bits.running:
            self.send(self.dev.turn_off)
        else:
            self.send(self.dev.turn_on)

    @QtCore.pyqtSlot()
    def process_pbtn_read_alarm_values(self):
        self.add_to_send_queue(self.dev.query_alarm_values_and_units)
        self.add_to_send_queue("signal_GUI_alarm_values_update")
        self.process_send_queue()

    @QtCore.pyqtSlot()
    def process_pbtn_read_PID_values(self):
        self.add_to_send_queue(self.dev.query_PID_values)
        self.add_to_send_queue("signal_GUI_PID_values_update")
        self.process_send_queue()

    @QtCore.pyqtSlot()
    def send_setpoint_from_textbox(self):
        try:
            setpoint = float(self.send_setpoint.text())
        except (TypeError, ValueError):
            setpoint = 22.0
        except:
            raise

        setpoint = max(setpoint, self.dev.min_setpoint_degC)
        setpoint = min(setpoint, self.dev.max_setpoint_degC)
        self.send_setpoint.setText("%.1f" % setpoint)

        self.send(self.dev.send_setpoint, setpoint)

    # --------------------------------------------------------------------------
    #   connect_signals_to_slots
    # --------------------------------------------------------------------------

    def connect_signals_to_slots(self):
        self.pbtn_on.clicked.connect(self.process_pbtn_on)
        self.pbtn_read_alarm_values.clicked.connect(
            self.process_pbtn_read_alarm_values
        )
        self.pbtn_read_PID_values.clicked.connect(
            self.process_pbtn_read_PID_values
        )
        self.send_setpoint.editingFinished.connect(
            self.send_setpoint_from_textbox
        )

        self.signal_GUI_alarm_values_update.connect(
            self.update_GUI_alarm_values
        )
        self.signal_GUI_PID_values_update.connect(self.update_GUI_PID_values)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Multithreaded PyQt5 GUI to interface with a Thermo Scientific ThermoFlex
recirculating chiller.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "07-07-2020"  # 0.0.1 was stamped 14-09-2018
__version__ = "0.0.5"  # 0.0.1 corresponds to prototype 1.0.0

import sys
from pathlib import Path

from PyQt5 import QtCore, QtGui
from PyQt5 import QtWidgets as QtWid

from dvg_utils.dvg_pyqt_controls import SS_TEXTBOX_READ_ONLY

from dvg_devices.ThermoFlex_chiller_protocol_RS232 import ThermoFlex_chiller
from dvg_devices.ThermoFlex_chiller_qdev import ThermoFlex_chiller_qdev

# ------------------------------------------------------------------------------
#   MainWindow
# ------------------------------------------------------------------------------


class MainWindow(QtWid.QWidget):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        self.setGeometry(40, 60, 0, 0)
        self.setWindowTitle("ThermoFlex chiller control")

        # Top grid
        self.lbl_title = QtWid.QLabel(
            "ThermoFlex chiller control",
            font=QtGui.QFont("Palatino", 14, weight=QtGui.QFont.Bold),
        )
        self.pbtn_exit = QtWid.QPushButton("Exit")
        self.pbtn_exit.clicked.connect(self.close)
        self.pbtn_exit.setMinimumHeight(30)

        grid_top = QtWid.QGridLayout()
        grid_top.addWidget(self.lbl_title, 0, 0)
        grid_top.addWidget(self.pbtn_exit, 0, 1, QtCore.Qt.AlignRight)

        # Round up full window
        vbox = QtWid.QVBoxLayout(self)
        vbox.addLayout(grid_top)
        vbox.addLayout(chiller_qdev.hbly_GUI)
        vbox.addStretch(1)


# ------------------------------------------------------------------------------
#   about_to_quit
# ------------------------------------------------------------------------------


def about_to_quit():
    print("About to quit")
    app.processEvents()
    chiller_qdev.quit()
    chiller.close()


# ------------------------------------------------------------------------------
#   Main
# ------------------------------------------------------------------------------

if __name__ == "__main__":
    # Temperature setpoint limits in software, not on a hardware level
    MIN_SETPOINT_DEG_C = 10  # [deg C]
    MAX_SETPOINT_DEG_C = 40  # [deg C]

    # Config file containing COM port address
    PATH_CONFIG = Path("config/port_chiller.txt")

    # The state of the chiller is polled with this time interval
    UPDATE_INTERVAL_MS = 1000  # [ms]

    # --------------------------------------------------------------------------
    #   Connect to ThermoFlex chiller
    # --------------------------------------------------------------------------

    chiller = ThermoFlex_chiller(
        min_setpoint_degC=MIN_SETPOINT_DEG_C,
        max_setpoint_degC=MAX_SETPOINT_DEG_C,
        name="chiller",
    )
    if chiller.auto_connect(PATH_CONFIG):
        chiller.begin()

    # --------------------------------------------------------------------------
    #   Create application
    # --------------------------------------------------------------------------
    QtCore.QThread.currentThread().setObjectName("MAIN")  # For DEBUG info

    app = 0  # Work-around for kernel crash when using Spyder IDE
    app = QtWid.QApplication(sys.argv)
    app.setFont(QtGui.QFont("Arial", 9))
    app.setStyleSheet(SS_TEXTBOX_READ_ONLY)
    app.aboutToQuit.connect(about_to_quit)

    # --------------------------------------------------------------------------
    #   Set up communication threads for the chiller
    # --------------------------------------------------------------------------

    chiller_qdev = ThermoFlex_chiller_qdev(chiller, UPDATE_INTERVAL_MS)
    chiller_qdev.start()

    # --------------------------------------------------------------------------
    #   Start the main GUI event loop
    # --------------------------------------------------------------------------

    window = MainWindow()
    window.show()
    sys.exit(app.exec_())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Multithreaded PyQt5 GUI to interface with a a Keysight (former HP or Agilent)
34970A/34972A data acquisition/switch unit.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "07-07-2020"  # 0.0.1 was stamped 14-09-2018
__version__ = "0.0.5"  # 0.0.1 corresponds to prototype 1.0.0
# pylint: disable=bare-except

import sys
import visa
import matplotlib.pyplot as plt
import numpy as np

from PyQt5 import QtCore, QtGui
from PyQt5 import QtWidgets as QtWid
from PyQt5.QtCore import QDateTime
import pyqtgraph as pg

from dvg_utils.dvg_pyqt_controls import (
    create_Toggle_button,
    SS_TEXTBOX_READ_ONLY,
    SS_GROUP,
)
from dvg_utils.dvg_pyqt_charthistory import ChartHistory
from dvg_utils.dvg_pyqt_filelogger import FileLogger

from dvg_devices.Keysight_3497xA_protocol_SCPI import Keysight_3497xA
from dvg_devices.Keysight_3497xA_qdev import Keysight_3497xA_qdev, INFINITY_CAP

# ------------------------------------------------------------------------------
#   MainWindow
# ------------------------------------------------------------------------------


class MainWindow(QtWid.QWidget):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        self.setGeometry(600, 120, 1200, 600)
        self.setWindowTitle("Keysight 3497xA control")

        # ----------------------------------------------------------------------
        #   Top grid
        # ----------------------------------------------------------------------

        self.qlbl_title = QtWid.QLabel(
            "Keysight 3497xA control",
            font=QtGui.QFont("Palatino", 14, weight=QtGui.QFont.Bold),
        )
        self.qlbl_cur_date_time = QtWid.QLabel("00-00-0000    00:00:00")
        self.qpbt_record = create_Toggle_button(
            "Click to start recording to file", minimumHeight=40
        )
        self.qpbt_record.setMinimumWidth(400)

        self.qpbt_exit = QtWid.QPushButton("Exit")
        self.qpbt_exit.clicked.connect(self.close)
        self.qpbt_exit.setMinimumHeight(30)

        grid_top = QtWid.QGridLayout()
        # fmt: off
        grid_top.addWidget(self.qlbl_title        , 0, 0, QtCore.Qt.AlignCenter)
        grid_top.addWidget(self.qpbt_exit         , 0, 2, QtCore.Qt.AlignRight)
        grid_top.addWidget(self.qlbl_cur_date_time, 1, 0, QtCore.Qt.AlignCenter)
        grid_top.addWidget(self.qpbt_record       , 2, 0, QtCore.Qt.AlignCenter)
        # fmt: on
        grid_top.setColumnMinimumWidth(0, 420)
        grid_top.setColumnStretch(1, 1)

        # ----------------------------------------------------------------------
        #   Chart: Mux readings
        # ----------------------------------------------------------------------

        # GraphicsWindow
        self.gw_mux = pg.GraphicsWindow()
        self.gw_mux.setBackground([20, 20, 20])

        # PlotItem
        self.pi_mux = self.gw_mux.addPlot()
        self.pi_mux.setTitle('<span style="font-size:12pt">Mux readings</span>')
        self.pi_mux.setLabel(
            "bottom", '<span style="font-size:12pt">history (min)</span>'
        )
        self.pi_mux.setLabel(
            "left", '<span style="font-size:12pt">misc. units</span>'
        )
        self.pi_mux.showGrid(x=1, y=1)
        self.pi_mux.setMenuEnabled(True)
        self.pi_mux.enableAutoRange(axis=pg.ViewBox.XAxis, enable=False)
        self.pi_mux.enableAutoRange(axis=pg.ViewBox.YAxis, enable=True)
        self.pi_mux.setAutoVisible(y=True)

        # Viewbox properties for the legend
        vb = self.gw_mux.addViewBox(enableMenu=False)
        vb.setMaximumWidth(80)

        # Legend
        self.legend = pg.LegendItem()
        self.legend.setParentItem(vb)
        self.legend.anchor((0, 0), (0, 0), offset=(1, 10))
        self.legend.setFixedWidth(75)
        self.legend.setScale(1)

        # ----------------------------------------------------------------------
        #   Show curves selection
        # ----------------------------------------------------------------------

        qgrp_show_curves = QtWid.QGroupBox("Show")
        qgrp_show_curves.setStyleSheet(SS_GROUP)
        self.grid_show_curves = QtWid.QGridLayout()
        self.grid_show_curves.setVerticalSpacing(0)
        qgrp_show_curves.setLayout(self.grid_show_curves)

        # ----------------------------------------------------------------------
        #   Chart history time range selection
        # ----------------------------------------------------------------------

        p = {"maximumWidth": 70}
        self.qpbt_history_1 = QtWid.QPushButton("00:30", **p)
        self.qpbt_history_2 = QtWid.QPushButton("01:00", **p)
        self.qpbt_history_3 = QtWid.QPushButton("03:00", **p)
        self.qpbt_history_4 = QtWid.QPushButton("05:00", **p)
        self.qpbt_history_5 = QtWid.QPushButton("10:00", **p)
        self.qpbt_history_6 = QtWid.QPushButton("30:00", **p)

        self.qpbt_history_clear = QtWid.QPushButton("clear", **p)
        self.qpbt_history_clear.clicked.connect(self.clear_all_charts)

        grid = QtWid.QGridLayout()
        grid.setVerticalSpacing(0)
        grid.addWidget(self.qpbt_history_1, 0, 0)
        grid.addWidget(self.qpbt_history_2, 1, 0)
        grid.addWidget(self.qpbt_history_3, 2, 0)
        grid.addWidget(self.qpbt_history_4, 3, 0)
        grid.addWidget(self.qpbt_history_5, 4, 0)
        grid.addWidget(self.qpbt_history_6, 5, 0)
        grid.addWidget(self.qpbt_history_clear, 6, 0)

        qgrp_history = QtWid.QGroupBox("History")
        qgrp_history.setStyleSheet(SS_GROUP)
        qgrp_history.setLayout(grid)

        # ----------------------------------------------------------------------
        #   Bottom grid
        # ----------------------------------------------------------------------

        vbox1 = QtWid.QVBoxLayout()
        vbox1.addWidget(
            qgrp_show_curves, stretch=0, alignment=QtCore.Qt.AlignTop
        )
        vbox1.addWidget(qgrp_history, stretch=0, alignment=QtCore.Qt.AlignTop)
        vbox1.addStretch(1)

        hbox1 = QtWid.QHBoxLayout()
        hbox1.addWidget(mux_qdev.qgrp, stretch=0, alignment=QtCore.Qt.AlignTop)
        hbox1.addWidget(self.gw_mux, stretch=1)
        hbox1.addLayout(vbox1)

        # ----------------------------------------------------------------------
        #   Round up full window
        # ----------------------------------------------------------------------

        vbox = QtWid.QVBoxLayout(self)
        vbox.addLayout(grid_top)
        vbox.addLayout(hbox1)
        vbox.addStretch(1)

    @QtCore.pyqtSlot()
    def clear_all_charts(self):
        str_msg = "Are you sure you want to clear all charts?"
        reply = QtWid.QMessageBox.warning(
            self,
            "Clear charts",
            str_msg,
            QtWid.QMessageBox.Yes | QtWid.QMessageBox.No,
            QtWid.QMessageBox.No,
        )

        if reply == QtWid.QMessageBox.Yes:
            [CH.clear() for CH in self.CHs_mux]


# ------------------------------------------------------------------------------
#   update_GUI
# ------------------------------------------------------------------------------


@QtCore.pyqtSlot()
def update_GUI():
    cur_date_time = QDateTime.currentDateTime()
    window.qlbl_cur_date_time.setText(
        cur_date_time.toString("dd-MM-yyyy")
        + "    "
        + cur_date_time.toString("HH:mm:ss")
    )

    # Update curves
    [CH.update_curve() for CH in window.CHs_mux]

    # Show or hide curve depending on checkbox
    for i in range(N_channels):
        window.CHs_mux[i].curve.setVisible(
            window.chkbs_show_curves[i].isChecked()
        )


# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------


@QtCore.pyqtSlot()
def process_qpbt_history_1():
    change_history_axes(
        time_axis_factor=1e3,  # transform [msec] to [sec]
        time_axis_range=-30,  # [sec]
        time_axis_label='<span style="font-size:12pt">history (sec)</span>',
    )


@QtCore.pyqtSlot()
def process_qpbt_history_2():
    change_history_axes(
        time_axis_factor=1e3,  # transform [msec] to [sec]
        time_axis_range=-60,  # [sec]
        time_axis_label='<span style="font-size:12pt">history (sec)</span>',
    )


@QtCore.pyqtSlot()
def process_qpbt_history_3():
    change_history_axes(
        time_axis_factor=60e3,  # transform [msec] to [min]
        time_axis_range=-3,  # [min]
        time_axis_label='<span style="font-size:12pt">history (min)</span>',
    )


@QtCore.pyqtSlot()
def process_qpbt_history_4():
    change_history_axes(
        time_axis_factor=60e3,  # transform [msec] to [min]
        time_axis_range=-5,  # [min]
        time_axis_label='<span style="font-size:12pt">history (min)</span>',
    )


@QtCore.pyqtSlot()
def process_qpbt_history_5():
    change_history_axes(
        time_axis_factor=60e3,  # transform [msec] to [min]
        time_axis_range=-10,  # [min]
        time_axis_label='<span style="font-size:12pt">history (min)</span>',
    )


@QtCore.pyqtSlot()
def process_qpbt_history_6():
    change_history_axes(
        time_axis_factor=60e3,  # transform [msec] to [min]
        time_axis_range=-30,  # [min]
        time_axis_label='<span style="font-size:12pt">history (min)</span>',
    )


def change_history_axes(time_axis_factor, time_axis_range, time_axis_label):
    window.pi_mux.setXRange(time_axis_range, 0)
    window.pi_mux.setLabel("bottom", time_axis_label)

    for i in range(N_channels):
        window.CHs_mux[i].x_axis_divisor = time_axis_factor


@QtCore.pyqtSlot()
def process_qpbt_show_all_curves():
    # First: if any curve is hidden --> show all
    # Second: if all curves are shown --> hide all

    any_hidden = False
    for i in range(N_channels):
        if not window.chkbs_show_curves[i].isChecked():
            window.chkbs_show_curves[i].setChecked(True)
            any_hidden = True

    if not any_hidden:
        for i in range(N_channels):
            window.chkbs_show_curves[i].setChecked(False)


@QtCore.pyqtSlot()
def process_qpbt_record():
    if window.qpbt_record.isChecked():
        file_logger.starting = True
    else:
        file_logger.stopping = True


@QtCore.pyqtSlot(str)
def set_text_qpbt_record(text_str):
    window.qpbt_record.setText(text_str)


# ------------------------------------------------------------------------------
#   about_to_quit
# ------------------------------------------------------------------------------


def about_to_quit():
    print("About to quit")
    app.processEvents()
    mux_qdev.quit()
    file_logger.close_log()

    try:
        mux.close()
    except:
        pass
    try:
        rm.close()
    except:
        pass


# ------------------------------------------------------------------------------
#   DAQ_postprocess_MUX_scan_function
# ------------------------------------------------------------------------------


def DAQ_postprocess_MUX_scan_function():
    """Will be called during an 'worker_DAQ' update, after a mux scan has been
    performed. We use it to parse out the scan readings into separate variables
    and log it to file.
    """
    cur_date_time = QDateTime.currentDateTime()
    epoch_time = cur_date_time.toMSecsSinceEpoch()

    # DEBUG info
    # dprint("thread: %s" % QtCore.QThread.currentThread().objectName())

    if mux_qdev.is_MUX_scanning:
        readings = mux.state.readings
        for i in range(N_channels):
            if readings[i] > INFINITY_CAP:
                readings[i] = np.nan
    else:
        readings = [np.nan] * N_channels
        mux.state.readings = readings

    # Add readings to charts
    for i in range(N_channels):
        window.CHs_mux[i].add_new_reading(epoch_time, readings[i])

    # ----------------------------------------------------------------------
    #   Logging to file
    # ----------------------------------------------------------------------

    if file_logger.starting:
        fn_log = (
            "d:/data/mux_" + cur_date_time.toString("yyMMdd_HHmmss") + ".txt"
        )
        if file_logger.create_log(epoch_time, fn_log, mode="w"):
            file_logger.signal_set_recording_text.emit(
                "Recording to file: " + fn_log
            )

            # Header
            file_logger.write("time[s]\t")
            for i in range(N_channels - 1):
                file_logger.write(
                    "CH%s\t" % mux.state.all_scan_list_channels[i]
                )
            file_logger.write("CH%s\n" % mux.state.all_scan_list_channels[-1])

    if file_logger.stopping:
        file_logger.signal_set_recording_text.emit(
            "Click to start recording to file"
        )
        file_logger.close_log()

    if file_logger.is_recording:
        log_elapsed_time = (epoch_time - file_logger.start_time) / 1e3  # [sec]

        # Add new data to the log
        file_logger.write("%.3f" % log_elapsed_time)
        for i in range(N_channels):
            if len(mux.state.readings) <= i:
                file_logger.write("\t%.5e" % np.nan)
            else:
                file_logger.write("\t%.5e" % mux.state.readings[i])
        file_logger.write("\n")


# ------------------------------------------------------------------------------
#   Main
# ------------------------------------------------------------------------------

if __name__ == "__main__":
    # VISA address of the Keysight 3497xA data acquisition/switch unit
    # containing a multiplexer plug-in module. Hence, we simply call this device
    # a 'mux'.
    # MUX_VISA_ADDRESS = "USB0::0x0957::0x2007::MY49018071::INSTR"
    MUX_VISA_ADDRESS = "GPIB0::9::INSTR"

    # A scan will be performed by the mux every N milliseconds
    MUX_SCANNING_INTERVAL_MS = 1000  # [ms]

    # Chart history (CH) buffer sizes in [samples].
    # Multiply this with the corresponding SCANNING_INTERVAL constants to get
    # the history size in time.
    CH_SAMPLES_MUX = 1800

    # The chart will be updated at this interval
    UPDATE_INTERVAL_GUI = 1000  # [ms]

    # SCPI commands to be send to the 3497xA to set up the scan cycle.
    """
    scan_list = "(@301:310)"
    MUX_SCPI_COMMANDS = [
                "rout:open %s" % scan_list,
                "conf:temp TC,J,%s" % scan_list,
                "unit:temp C,%s" % scan_list,
                "sens:temp:tran:tc:rjun:type INT,%s" % scan_list,
                "sens:temp:tran:tc:check ON,%s" % scan_list,
                "sens:temp:nplc 1,%s" % scan_list,
                "rout:scan %s" % scan_list]
    """
    scan_list = "(@101)"
    MUX_SCPI_COMMANDS = [
        "rout:open %s" % scan_list,
        "conf:res 1e5,%s" % scan_list,
        "sens:res:nplc 1,%s" % scan_list,
        "rout:scan %s" % scan_list,
    ]

    # --------------------------------------------------------------------------
    #   Connect to Keysight 3497xA (mux)
    # --------------------------------------------------------------------------

    rm = visa.ResourceManager()

    mux = Keysight_3497xA(MUX_VISA_ADDRESS, "MUX")
    if mux.connect(rm):
        mux.begin(MUX_SCPI_COMMANDS)

    # --------------------------------------------------------------------------
    #   Create application
    # --------------------------------------------------------------------------
    QtCore.QThread.currentThread().setObjectName("MAIN")  # For DEBUG info

    app = 0  # Work-around for kernel crash when using Spyder IDE
    app = QtWid.QApplication(sys.argv)
    app.setFont(QtGui.QFont("Arial", 9))
    app.setStyleSheet(SS_TEXTBOX_READ_ONLY)
    app.aboutToQuit.connect(about_to_quit)

    # Create PyQt GUI interfaces and communication threads per 3497xA
    mux_qdev = Keysight_3497xA_qdev(
        dev=mux,
        DAQ_interval_ms=MUX_SCANNING_INTERVAL_MS,
        DAQ_postprocess_MUX_scan_function=DAQ_postprocess_MUX_scan_function,
    )
    mux_qdev.set_table_readings_format("%.5e")
    mux_qdev.qgrp.setFixedWidth(420)

    # Create window
    window = MainWindow()

    # --------------------------------------------------------------------------
    #   Create pens and chart histories depending on the number of scan channels
    # --------------------------------------------------------------------------

    N_channels = len(mux.state.all_scan_list_channels)

    # Pen styles for plotting
    PENS = [None] * N_channels
    cm = plt.get_cmap("gist_rainbow")
    params = {"width": 2}
    for i in range(N_channels):
        color = cm(1.0 * i / N_channels)  # color will now be an RGBA tuple
        color = np.array(color) * 255
        PENS[i] = pg.mkPen(color=color, **params)

    # Create Chart Histories (CH) and PlotDataItems and link them together
    # Also add legend entries
    window.CHs_mux = [None] * N_channels
    window.chkbs_show_curves = [None] * N_channels
    for i in range(N_channels):
        window.CHs_mux[i] = ChartHistory(
            CH_SAMPLES_MUX, window.pi_mux.plot(pen=PENS[i])
        )
        window.legend.addItem(
            window.CHs_mux[i].curve, name=mux.state.all_scan_list_channels[i]
        )

        # Add checkboxes for showing the curves
        window.chkbs_show_curves[i] = QtWid.QCheckBox(
            parent=window,
            text=mux.state.all_scan_list_channels[i],
            checked=True,
        )
        window.grid_show_curves.addWidget(window.chkbs_show_curves[i], i, 0)

    window.qpbt_show_all_curves = QtWid.QPushButton("toggle", maximumWidth=70)
    window.qpbt_show_all_curves.clicked.connect(process_qpbt_show_all_curves)
    window.grid_show_curves.addWidget(
        window.qpbt_show_all_curves, N_channels, 0
    )

    # --------------------------------------------------------------------------
    #   File logger
    # --------------------------------------------------------------------------

    file_logger = FileLogger()
    file_logger.signal_set_recording_text.connect(set_text_qpbt_record)

    # --------------------------------------------------------------------------
    #   Start threads
    # --------------------------------------------------------------------------

    mux_qdev.start(DAQ_priority=QtCore.QThread.TimeCriticalPriority)

    # --------------------------------------------------------------------------
    #   Connect remaining signals from GUI
    # --------------------------------------------------------------------------

    window.qpbt_history_1.clicked.connect(process_qpbt_history_1)
    window.qpbt_history_2.clicked.connect(process_qpbt_history_2)
    window.qpbt_history_3.clicked.connect(process_qpbt_history_3)
    window.qpbt_history_4.clicked.connect(process_qpbt_history_4)
    window.qpbt_history_5.clicked.connect(process_qpbt_history_5)
    window.qpbt_history_6.clicked.connect(process_qpbt_history_6)
    window.qpbt_record.clicked.connect(process_qpbt_record)

    # --------------------------------------------------------------------------
    #   Set up timers
    # --------------------------------------------------------------------------

    timer_GUI = QtCore.QTimer()
    timer_GUI.timeout.connect(update_GUI)
    timer_GUI.start(UPDATE_INTERVAL_GUI)

    # --------------------------------------------------------------------------
    #   Start the main GUI event loop
    # --------------------------------------------------------------------------

    # Init the time axis of the strip charts
    process_qpbt_history_3()

    window.show()
    sys.exit(app.exec_())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""RS232 function library for Bronkhorst mass flow controllers (MFC) using the
FLOW-BUS protocol.

Only the ASCII version is supported, not the enhanced binary version. This
library sends and receives messages to/from all nodes (code 80), hence just one
MFC is assumed per port.

When this module is directly run from the terminal a demo will be shown.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "06-07-2020"  # 0.0.1 was stamped 25-07-2018
__version__ = "0.0.5"  # 0.0.1 corresponds to prototype 1.0.0
# pylint: disable=bare-except, try-except-raise

import sys
import struct
from pathlib import Path

import serial
import serial.tools.list_ports

from dvg_debug_functions import print_fancy_traceback as pft

# Serial settings
RS232_BAUDRATE = 38400  # Baudrate according to the manual
RS232_TIMEOUT = 0.1  # [sec]


class Bronkhorst_MFC:
    class State:
        # Container for the process and measurement variables
        setpoint = None  # Setpoint read out of the MFC   [ln/min]
        flow_rate = None  # Flow rate measured by the MFC [ln/min]

    # --------------------------------------------------------------------------
    #   __init__
    # --------------------------------------------------------------------------

    def __init__(self, name="MFC"):
        self.ser = None  # serial.Serial device instance
        self.name = name
        self.serial_str = None  # Serial number of the MFC
        self.model_str = None  # Model of the MFC
        self.fluid_name = None  # Fluid for which the MFC is calibrated
        self.max_flow_rate = None  # Max. capacity [ln/min]

        # Is the connection to the device alive?
        self.is_alive = False

        # Container for the process and measurement variables
        self.state = self.State()

    # --------------------------------------------------------------------------
    #   close
    # --------------------------------------------------------------------------

    def close(self, ignore_exceptions=False):
        """Close the serial port
        """
        if self.ser is not None:
            try:
                self.ser.close()
            except:
                if ignore_exceptions:
                    pass
                else:
                    raise

        self.is_alive = False

    # --------------------------------------------------------------------------
    #   connect_at_port
    # --------------------------------------------------------------------------

    def connect_at_port(
        self, port_str, match_serial_str=None, print_trying_message=True
    ):
        """Open the port at address 'port_str' and try to establish a
        connection. A query for the Bronkhorst serial number is send over the
        port. If it gives the proper response (and optionally has a matching
        serial number) it must be the Bronkhorst MFC we're looking for.

        Args:
            port_str (str): Serial port address to open
            match_serial_str (str, optional): Serial string of the MFC to
                establish a connection to. When empty or None then any MFC is
                accepted. Defaults to None.
            print_trying_message (bool, optional): When True then a 'trying to
                open' message is printed to the terminal. Defaults to True.

        Returns: True if successful, False otherwise.
        """
        self.is_alive = False

        if match_serial_str == "":
            match_serial_str = None
        if print_trying_message:
            if match_serial_str is None:
                print("Connect to: Bronkhorst MFC")
            else:
                print(
                    "Connect to: Bronkhorst MFC, serial %s" % match_serial_str
                )

        print("  @ %-5s: " % port_str, end="")
        try:
            # Open the serial port
            self.ser = serial.Serial(
                port=port_str,
                baudrate=RS232_BAUDRATE,
                timeout=RS232_TIMEOUT,
                write_timeout=RS232_TIMEOUT,
            )
        except serial.SerialException:
            print("Could not open port")
            return False
        except:
            raise

        try:
            # Query the serial number string.
            # NOTE: this function can finish okay and return False as indication
            # that the device on the serial port gives /a/ reply but it is not a
            # proper reply you would except from a Bronkhorst MFC. In other
            # words: the device replies but it is not a Bronkhorst MFC.
            self.is_alive = True
            success = self.query_serial_str()
        except:
            print("Communication error")
            if self.ser is not None:
                self.ser.close()
            self.is_alive = False
            return False

        if success:
            print("serial %s: " % self.serial_str, end="")
            if match_serial_str is None:
                # Found any Bronkhorst MFC device
                print("Success!\n")
                self.is_alive = True
                return True
            elif self.serial_str.lower() == match_serial_str.lower():
                # Found the Bronkhorst MFC with matching serial
                print("Success!\n")
                self.is_alive = True
                return True

        print("Wrong or no device")
        if self.ser is not None:
            self.ser.close()
        self.is_alive = False
        return False

    # --------------------------------------------------------------------------
    #   scan_ports
    # --------------------------------------------------------------------------

    def scan_ports(self, match_serial_str=None):
        """Scan over all serial ports and try to establish a connection. A query
        for the Bronkhorst serial number is send over all ports. The port that
        gives the proper response (and optionally has a matching serial number)
        must be the Bronkhorst MFC we're looking for.

        Args:
            match_serial_str (str, optional): Serial string of the MFC to
                establish a connection to. When empty or None then any MFC is
                accepted. Defaults to None.

        Returns: True if successful, False otherwise.
        """
        if match_serial_str == "":
            match_serial_str = None
        if match_serial_str is None:
            print("Scanning ports for any Bronkhorst MFC")
        else:
            print(
                (
                    "Scanning ports for a Bronkhorst MFC with\n"
                    "serial number '%s'"
                )
                % match_serial_str
            )

        # Ports is a list of tuples
        ports = list(serial.tools.list_ports.comports())
        for p in ports:
            port_str = p[0]
            if self.connect_at_port(port_str, match_serial_str, False):
                return True
            else:
                continue

        # Scanned over all the ports without finding a match
        print("\n  ERROR: Device not found")
        return False

    # --------------------------------------------------------------------------
    #   auto_connect
    # --------------------------------------------------------------------------

    def auto_connect(self, path_config, match_serial_str=None):
        """TO DO: write explaination
        """
        # Try to open the config file containing the port to open. Do not panic
        # if the file does not exist or cannot be read. We will then scan over
        # all ports as alternative.
        port_str = read_port_config_file(path_config)

        # If the config file was read successfully then we can try to open the
        # port listed in the config file and connect to the device.
        if port_str is not None:
            success = self.connect_at_port(port_str, match_serial_str)
        else:
            success = False

        # Check if we failed establishing a connection
        if not success:
            # Now scan over all ports and try to connect to the device
            success = self.scan_ports(match_serial_str)
            if success:
                # Store the result of a successful connection after a port scan
                # in the config file. Do not panic if we cannot create the
                # config file.
                write_port_config_file(path_config, self.ser.portstr)

        return success

    # --------------------------------------------------------------------------
    #   begin
    # --------------------------------------------------------------------------

    def begin(self):
        """This function should run directly after having established a
        connection to the device.

        Query the model, fluid and the maximum mass flow rate of the MFC
        and store these in the class member 'state's. The max flow rate is
        mandatory to be known, because it is used to set and read the setpoint,
        and to read the flow rate.

        Returns: True if successful, False otherwise.
        """
        success = False
        success &= self.query_model_str()
        success &= self.query_fluid_name()
        success &= self.query_max_flow_rate()
        success &= self.query_setpoint()
        success &= self.query_flow_rate()
        return success

    # --------------------------------------------------------------------------
    #   query
    # --------------------------------------------------------------------------

    def query(self, msg_str):
        """Send a command to the serial device and subsequently read the reply.

        Args:
            msg_str (str): Message to be sent to the serial device.

        Returns:
            success (bool): True if successful, False otherwise.
            ans_str (str): Reply received from the device. None if unsuccessful.
        """
        success = False
        ans_str = None

        if not self.is_alive:
            print("ERROR: Device is not connected yet or already closed.")
        else:
            try:
                # Send command string to the device as bytes
                self.ser.write(msg_str.replace(" ", "").encode())
            except (
                serial.SerialTimeoutException,
                serial.SerialException,
            ) as err:
                # Print error and struggle on
                pft(err, 3)
            except:
                raise
            else:
                try:
                    # Read all bytes in the line that is terminated with a
                    # newline character or until time-out has occured
                    ans_bytes = self.ser.readline()
                except (
                    serial.SerialTimeoutException,
                    serial.SerialException,
                ) as err:
                    pft(err, 3)
                except:
                    raise
                else:
                    # Convert bytes into string and remove termination chars and
                    # spaces
                    ans_str = ans_bytes.decode().strip()
                    success = True

        return [success, ans_str]

    # --------------------------------------------------------------------------
    #   query_serial_str
    # --------------------------------------------------------------------------

    def query_serial_str(self):
        """Query the serial number and store it in the class member 'serial_str'

        Returns: True if successful, False otherwise.
        """
        [success, ans] = self.query(":07 80 04 71 63 71 63 00\r\n")
        if success and ans[3:13] == "8002716300":
            self.serial_str = bytearray.fromhex(ans[13:-2]).decode()
            return True
        else:
            self.serial_str = None
            return False

    # --------------------------------------------------------------------------
    #   query_model_str
    # --------------------------------------------------------------------------

    def query_model_str(self):
        """Query the model name of the MFC and store it in the class member 'state'.

        Returns: True if successful, False otherwise.
        """
        [success, ans] = self.query(":07 80 04 71 62 71 62 00\r\n")
        if success:
            self.model_str = bytearray.fromhex(ans[13:-2]).decode()
            return True
        else:
            self.model_str = None
            return False

    # --------------------------------------------------------------------------
    #   query_fluid_name
    # --------------------------------------------------------------------------

    def query_fluid_name(self):
        """Query the fluid name that the MFC is calibrated for and store it in
        the class member 'state'.

        Returns: True if successful, False otherwise.
        """
        [success, ans] = self.query(":07 80 04 01 71 01 71 0A\r\n")
        if success:
            self.fluid_name = bytearray.fromhex(ans[13:-2]).decode()
            return True
        else:
            self.fluid_name = None
            return False

    # --------------------------------------------------------------------------
    #   query_max_flow_rate
    # --------------------------------------------------------------------------

    def query_max_flow_rate(self):
        """Query the maximum mass flow rate in [ln/min] of the MFC and store it
        in the class member 'state'. This value is mandatory to be known, because it is
        used to set and read the setpoint, and to read the flow rate.

        Returns: True if successful, False otherwise.
        """
        [success, ans] = self.query(":06 80 04 01 4D 01 4D\r\n")
        if success:
            self.max_flow_rate = hex_to_32bit_IEEE754_float(ans[11:])
            return True
        else:
            self.max_flow_rate = None
            return False

    # --------------------------------------------------------------------------
    #   query_setpoint
    # --------------------------------------------------------------------------

    def query_setpoint(self):
        """Query the mass flow rate setpoint in [ln/min] set at the MFC and
        store it in the class member 'state' 'state'. Will be set to None if
        unsuccessful.

        Returns: True if successful, False otherwise.
        """
        [success, ans] = self.query(":06 80 04 01 21 01 21\r\n")
        if success:
            try:
                num = int(ans[-4:], 16)
            except ValueError:
                pass
            else:
                self.state.setpoint = num / 32000.0 * self.max_flow_rate
                return True

        self.state.setpoint = None
        return False

    # --------------------------------------------------------------------------
    #   query_flow_rate
    # --------------------------------------------------------------------------

    def query_flow_rate(self):
        """Query the mass flow rate in [ln/min] measured by the MFC and
        store it in the class member 'state'. Will be set to None if
        unsuccessful.

        Returns: True if successful, False otherwise.
        """
        [success, ans] = self.query(":06 80 04 01 21 01 20\r\n")
        if success:
            try:
                num = int(ans[-4:], 16)
            except ValueError:
                pass
            else:
                self.state.flow_rate = num / 32000.0 * self.max_flow_rate
                return True

        self.state.flow_rate = None
        return False

    # --------------------------------------------------------------------------
    #   send_setpoint
    # --------------------------------------------------------------------------

    def send_setpoint(self, setpoint):
        """Send a new mass flow rate setpoint in [ln/min] to the MFC.

        Args:
            setpoint (float): mass flow rate in [ln/min].

        Returns: True if successful, False otherwise.
        """
        try:
            setpoint = float(setpoint)
        except (TypeError, ValueError):
            setpoint = 0.0

        # Transform setpoint and limit
        setpoint = int(setpoint / self.max_flow_rate * 32000)
        setpoint = max(0, min(setpoint, 32000))

        [success, ans] = self.query(":06 80 01 01 21 %04x \r\n" % setpoint)
        if success and ans[5:].strip() == "000005":  # Also check status reply
            return True
        else:
            return False


# ------------------------------------------------------------------------------
#   hex_to_32bit_IEEE754_float
# ------------------------------------------------------------------------------


def hex_to_32bit_IEEE754_float(hex_str):
    """Transform a string containing a hexidecimal representation of a 32-bits
    IEEE754-formatted float value to a float
    """
    return (struct.unpack("f", struct.pack("i", int(hex_str, 16))))[0]


# -----------------------------------------------------------------------------
#   read_port_config_file
# -----------------------------------------------------------------------------


def read_port_config_file(filepath):
    """Try to open the config textfile containing the port to open. Do not panic
    if the file does not exist or cannot be read.

    Args:
        filepath (pathlib.Path): path to the config file,
            e.g. Path("config/port.txt")

    Returns: The port name string when the config file is read out successfully,
        None otherwise.
    """
    if isinstance(filepath, Path):
        if filepath.is_file():
            try:
                with filepath.open() as f:
                    port_str = f.readline().strip()
                return port_str
            except:
                pass  # Do not panic and remain silent

    return None


# -----------------------------------------------------------------------------
#   write_port_config_file
# -----------------------------------------------------------------------------


def write_port_config_file(filepath, port_str):
    """Try to write the port name string to the config textfile. Do not panic if
    the file cannot be created.

    Args:
        filepath (pathlib.Path): path to the config file,
            e.g. Path("config/port.txt")
        port_str (string): COM port string to save to file
    Returns: True when successful, False otherwise.
    """
    if isinstance(filepath, Path):
        if not filepath.parent.is_dir():
            # Subfolder does not exists yet. Create.
            try:
                filepath.parent.mkdir()
            except:
                pass  # Do not panic and remain silent

        try:
            # Write the config file
            filepath.write_text(port_str)
        except:
            pass  # Do not panic and remain silent
        else:
            return True

    return False


# ------------------------------------------------------------------------------
#   Main: Will show a demo when run from the terminal
# ------------------------------------------------------------------------------

if __name__ == "__main__":
    import os
    import time

    # Serial number of the Bronkhorst MFC to connect to.
    # Set to '' or None to connect to any Bronkhorst MFC.
    # SERIAL_MFC = "M16216843A"
    SERIAL_MFC = None

    # Path to the config textfile containing the (last used) RS232 port
    PATH_CONFIG = Path("config/port_Bronkhorst_MFC_1.txt")

    # Create connection to Bronkhorst MFC over RS232
    mfc = Bronkhorst_MFC()

    if mfc.auto_connect(PATH_CONFIG, SERIAL_MFC):
        mfc.begin()  # Retrieve necessary parameters
        print("  Serial  : %s" % mfc.serial_str)
        print("  Model   : %s" % mfc.model_str)
        print("  Fluid   : %s" % mfc.fluid_name)
        print("  Capacity: %.2f ln/min" % mfc.max_flow_rate)
    else:
        time.sleep(1)
        sys.exit(0)

    if os.name == "nt":
        import msvcrt

        running_Windows = True
        print("\nPress Q to quit.")
        print("Press S to enter new setpoint.")
    else:
        running_Windows = False
        print("\nPress Control + C to quit.")
        print("No other keyboard input possible because OS is not Windows.")

    # Prepare
    send_setpoint = 0.0
    do_send_setpoint = False

    mfc.query_setpoint()
    print("\nSetpoint       : %6.2f ln/min" % mfc.state.setpoint)

    # Loop
    done = False
    while not done:
        # Check if a new setpoint has to be send
        if do_send_setpoint:
            mfc.send_setpoint(send_setpoint)
            mfc.query_setpoint()
            print("\nSetpoint       : %6.2f ln/min" % mfc.state.setpoint)
            do_send_setpoint = False

        # Measure and report the flow rate
        mfc.query_flow_rate()
        print("\rMeas. flow rate: %6.2f ln/min" % mfc.state.flow_rate, end="")
        sys.stdout.flush()

        # Process keyboard input
        if running_Windows:
            if msvcrt.kbhit():
                key = msvcrt.getch()
                if key == b"q":
                    print("\nAre you sure you want to quit [y/n]?")
                    if msvcrt.getch() == b"y":
                        print("Quitting.")
                        done = True
                    else:
                        do_send_setpoint = True  # Esthestics
                elif key == b"s":
                    send_setpoint = input("\nEnter new setpoint [ln/min]: ")
                    do_send_setpoint = True

        # Slow down update period
        time.sleep(0.02)

    mfc.ser.close()
    time.sleep(1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "07-07-2020"  # 0.0.1 was stamped 15-09-2018
__version__ = "0.0.5"  # 0.0.1 corresponds to prototype 1.0.0
# pylint: disable=try-except-raise

import socket
import numpy as np

# ITS-90 resistance-temperature relation for PT100/PT1000
# R_t = R_0 * (1 + A*t + B*t^2 + C*(t-100)*t^3)
# R_t: resistance at T 'C   [Ohm]
# R_0: resistance at 0 'C   [Ohm]
# t  : temperature          ['C]
A = 3.9083e-3
B = -5.775e-7
C = -4.183e-12  # when below 0 'C, C = 0 when above 0 'C

# Acceptable temperature range
# fmt: off
T_MIN = -200    # ['C]
T_MAX = 800     # ['C]

# Acceptable resistance range, used to determine if a PT100 or PT1000
# probe is present
R_MIN = 18      # [Ohm]
R_MAX = 3760    # [Ohm]
# fmt: on

# Timeout on the socket communication
# Keep the timeout shorter than the scan rate of ~ 720 ms otherwise the method
# 'scan_4_wire_temperature' will break.
SOCKET_TIMEOUT = 0.5  # 0.5 [s]

DEBUG = False


class Picotech_PT104:
    class Eeprom:
        # Container for the PT-104 specific values retreived from it's memory
        # fmt: off
        serial     = None
        calib_date = None
        ch1_calib  = None
        ch2_calib  = None
        ch3_calib  = None
        ch4_calib  = None
        MAC        = None
        checksum   = None
        # fmt: on

    class State:
        # Container for the process and measurement variables
        # Resistance readings of channels 1 to 4 [Ohm]
        ch1_R = np.nan
        ch2_R = np.nan
        ch3_R = np.nan
        ch4_R = np.nan
        # Temperature readings of channels 1 to 4 ['C]
        ch1_T = np.nan
        ch2_T = np.nan
        ch3_T = np.nan
        ch4_T = np.nan

    # --------------------------------------------------------------------------
    #   __init__
    # --------------------------------------------------------------------------

    def __init__(self, name="PT104"):
        self.name = name
        self._ip_address = None
        self._port = None
        self._sock = None
        self._eeprom = self.Eeprom()

        # List corresponding to channels 1 to 4, where
        #    0: channel off
        #    1: channel on
        self._ENA_channels = [0, 0, 0, 0]

        # List corresponding to channels 1 to 4, where
        #    0: 1x gain
        #    1: 21x gain (for 375 Ohm range)
        self._gain_channels = [0, 0, 0, 0]

        # Resistance at 0 'C of channels 1 to 4 [Ohm]
        # For a PT100  probe this should be 100.000 Ohm
        # For a PT1000 probe this should be 1000.000 Ohm
        self.ch1_R_0 = 100
        self.ch2_R_0 = 100
        self.ch3_R_0 = 100
        self.ch4_R_0 = 100

        # Is the connection to the device alive?
        self.is_alive = False

        # Container for the measurement variables
        self.state = self.State()

    # --------------------------------------------------------------------------
    #   close
    # --------------------------------------------------------------------------

    def close(self):
        self._sock.close()
        self.is_alive = False

    # --------------------------------------------------------------------------
    #   connect
    # --------------------------------------------------------------------------

    def connect(self, ip_address="10.10.100.2", port=1234):
        """
        Returns: True if successful, False otherwise.
        """
        self._ip_address = ip_address
        self._port = port

        print("Connect to: PicoTech PT-104")
        print("  @ ip=%s:%i : " % (ip_address, port), end="")

        # Open UDP socket
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._sock.settimeout(SOCKET_TIMEOUT)  # timeout on commands

        # Try to acquire a lock to the PT-104
        success = self.lock()

        if success:
            print("Success!\n")
            self.is_alive = True
        else:
            print("FAILED!\n")
            self.is_alive = False

        return success

    # --------------------------------------------------------------------------
    #   begin
    # --------------------------------------------------------------------------

    def begin(self):
        """
        Returns: True if successful, False otherwise.
        """
        success = self.lock()
        success &= self.read_EEPROM()
        return success

    # --------------------------------------------------------------------------
    #   UDP_send
    # --------------------------------------------------------------------------

    def UDP_send(self, msg_bytes):
        """
        msg_bytes (bytes): Message to be sent over the UDP port.
        """
        self._sock.sendto(msg_bytes, (self._ip_address, self._port))

    # --------------------------------------------------------------------------
    #   UDP_recv
    # --------------------------------------------------------------------------

    def UDP_recv(self):
        """Receive one UDP packet at a time when available.

        Returns:
            success (bool):
                True if a packet was received successfully, False otherwise.

            ans_bytes:
                UDP packet received from the device. None if unsuccessful.
        """
        success = False
        ans_bytes = None

        try:
            ans_bytes = self._sock.recv(4096)
            success = True
        except socket.timeout:
            # print("ERROR: socket.recv() timed out in query()")
            pass  # Stay silent and continue
        except:
            raise

        return (success, ans_bytes)

    # --------------------------------------------------------------------------
    #   UDP_query_and_check
    # --------------------------------------------------------------------------

    def UDP_query_and_check(self, msg_bytes, check_ans_bytes):
        """
        Args:
            msg_bytes (bytes):
                Message to be sent over the UDP port.

        Returns:
            success (bool):
                True if successful, False otherwise.
            ans_bytes (bytes):
                Reply received from the device. None if unsuccessful.
        """
        self.UDP_send(msg_bytes)

        # Try up to 3 times at maximum to receive the respective response
        # UDP packet belonging to above sent UDP packet.
        for _i in range(3):
            [success, ans_bytes] = self.UDP_recv()
            if success:
                if ans_bytes[: len(check_ans_bytes)] == check_ans_bytes:
                    return (True, ans_bytes)
                else:
                    print("Failed %s: received %s" % (msg_bytes, ans_bytes))

        return (False, None)

    # --------------------------------------------------------------------------
    #   PT-104 functions
    #   Will return True when successful, False when failed
    # --------------------------------------------------------------------------

    def lock(self):
        success, _ans = self.UDP_query_and_check(b"lock\r", b"Lock Success")
        return success

    def set_mains_rejection_50Hz(self):
        success, _ans = self.UDP_query_and_check(
            bytes([0x30, 0x00]), b"Mains Changed"
        )
        return success

    def set_mains_rejection_60Hz(self):
        success, _ans = self.UDP_query_and_check(
            bytes([0x30, 0x01]), b"Mains Changed"
        )
        return success

    def keep_alive(self):
        success, _ans = self.UDP_query_and_check(bytes([0x34]), b"Alive")
        if not success:
            print("PT104 is not alive anymore.")
        return success

    def read_EEPROM(self):
        success, ans = self.UDP_query_and_check(bytes([0x32]), b"Eeprom")

        if success:
            # Parse
            # fmt: off
            ans        = ans[7:]   # Discard the first 7 bytes reading 'Eeprom='
            serial     = ans[19:29].decode("UTF8")
            calib_date = ans[29:37].decode("UTF8")
            ch1_calib  = int.from_bytes(ans[37:41], byteorder="little")
            ch2_calib  = int.from_bytes(ans[41:45], byteorder="little")
            ch3_calib  = int.from_bytes(ans[45:49], byteorder="little")
            ch4_calib  = int.from_bytes(ans[49:53], byteorder="little")
            MAC        = ":".join("%02x" % b for b in ans[53:59])
            checksum   = " ".join("0x%02x" % b for b in ans[126:128])

            self._eeprom.serial     = serial
            self._eeprom.calib_date = calib_date
            self._eeprom.ch1_calib  = ch1_calib
            self._eeprom.ch2_calib  = ch2_calib
            self._eeprom.ch3_calib  = ch3_calib
            self._eeprom.ch4_calib  = ch4_calib
            self._eeprom.MAC        = MAC
            self._eeprom.checksum   = checksum
            # fmt: on

            return True
        else:
            return False

    def start_conversion(self, ENA_channels=None, gain_channels=None):
        """
        Starts the continuous acquisition of measurements over channels 1 to 4
        ENA_channel is a list corresponding to channels 1 to 4, where
            0: channel off
            1: channel on
        gain_channels is a list corresponding to channels 1 to 4, where
            0: 1x gain
            1: 21x gain (for 375 Ohm range)
        """
        self._ENA_channels = (
            [1, 0, 0, 0] if ENA_channels is None else ENA_channels
        )
        self._gain_channels = (
            [1, 0, 0, 0] if gain_channels is None else gain_channels
        )

        data_byte = 0
        data_byte += self._ENA_channels[0]
        data_byte += self._ENA_channels[1] * 2
        data_byte += self._ENA_channels[2] * 4
        data_byte += self._ENA_channels[3] * 8
        data_byte += gain_channels[0] * 16
        data_byte += gain_channels[1] * 32
        data_byte += gain_channels[2] * 64
        data_byte += gain_channels[3] * 128

        success, _ans = self.UDP_query_and_check(
            bytes([0x31, data_byte]), b"Converting"
        )
        return success

    # --------------------------------------------------------------------------
    #   report_EEPROM
    # --------------------------------------------------------------------------

    def report_EEPROM(self):
        print("EEPROM")
        print("  serial    : %s" % self._eeprom.serial)
        print("  calib_date: %s" % self._eeprom.calib_date)
        print("  ch1_calib : %s" % self._eeprom.ch1_calib)
        print("  ch2_calib : %s" % self._eeprom.ch2_calib)
        print("  ch3_calib : %s" % self._eeprom.ch3_calib)
        print("  ch4_calib : %s" % self._eeprom.ch4_calib)
        print("  MAC       : %s" % self._eeprom.MAC)
        print("  checksum  : %s" % self._eeprom.checksum)

    # --------------------------------------------------------------------------
    #   scan_4_wire_temperature
    # --------------------------------------------------------------------------

    def scan_4_wire_temperature(self):
        """Reads the UDP port for any message, presumably the measurement of the
        channels reported by the PT104, after conversion has initiated.
        These readings are transformed into resistance (Ohm) using the
        calibration constants retreived from EEPROM, and again transformed to
        temperature ('C) using the ITS-90 resistance-temperature relation
        for PT100/PT1000. Four-wire measurements are assumed.

        Returns: True if successful, False otherwise.
        """

        ## Send keep alive signal. We care about the reply later.
        self.UDP_send(bytes([0x34]))

        (success, ans) = self.UDP_recv()
        while success:
            if ans[0] == 0 or ans[0] == 4 or ans[0] == 8 or ans[0] == 12:
                # Packet containing temperature reading

                ch = (
                    ans[0] / 4 + 1
                )  # Determine the channel number being reported
                a_0 = int.from_bytes(ans[1:5], byteorder="big")
                a_1 = int.from_bytes(ans[6:10], byteorder="big")
                a_2 = int.from_bytes(ans[11:15], byteorder="big")
                a_3 = int.from_bytes(ans[16:20], byteorder="big")

                if DEBUG:
                    print("CH %i" % ch)

                # fmt: off
                if   ch==1: calib = self._eeprom.ch1_calib; R_0 = self.ch1_R_0
                elif ch==2: calib = self._eeprom.ch2_calib; R_0 = self.ch2_R_0
                elif ch==3: calib = self._eeprom.ch3_calib; R_0 = self.ch3_R_0
                elif ch==4: calib = self._eeprom.ch4_calib; R_0 = self.ch4_R_0
                # fmt: on

                # Transform readings to resistance [Ohm]
                if (a_1 - a_0) == 0:
                    R_T = np.nan
                else:
                    R_T = (calib * (a_3 - a_2)) / (a_1 - a_0) / 1e6

                if np.isnan(R_T):
                    # No probe is present on the channel
                    T = np.nan
                elif (R_T < R_MIN) | (R_T > R_MAX):
                    # No probe is present on the channel
                    T = np.nan
                else:
                    # Tranform resistance to temperature ['C]
                    T = ITS90_Ohm_to_degC(R_0, R_T)

                    # Significant numbers + 1
                    T = np.round(T * 1e4) / 1e4

                # fmt: off
                if   ch == 1: self.state.ch1_R = R_T; self.state.ch1_T = T
                elif ch == 2: self.state.ch2_R = R_T; self.state.ch2_T = T
                elif ch == 3: self.state.ch3_R = R_T; self.state.ch3_T = T
                elif ch == 4: self.state.ch4_R = R_T; self.state.ch4_T = T
                # fmt: on

            elif ans[:5] == b"Alive":
                # Packet containing alive response. Stay silent.
                pass

            else:
                # Other packet?
                print("  %s" % ans)
                return False

            # Receive a possible next packet from the UDP in-buffer
            (success, ans) = self.UDP_recv()

        # No more packets
        return True


# ------------------------------------------------------------------------------
#   ITS90 transform functions
# ------------------------------------------------------------------------------


def ITS90_degC_to_Ohm(R_0, T):
    # ITS-90 resistance-temperature relation
    # R_T = R_0 * (1 + A*T + B*T^2 + C*(T-100)*T^3)
    # R_T: resistance at T 'C   [Ohm]
    # R_0: resistance at 0 'C   [Ohm]
    # T  : temperature          ['C]
    # A = 3.9083e-3
    # B = -5.775e-7
    # C = -4.183e-12  # when below 0 'C, C = 0 when above 0 'C
    return R_0 * (1 + A * T + B * T ** 2 + C * (T - 100) * T ** 3)


def ITS90_Ohm_to_degC(R_0, R_T):
    # ITS-90 resistance-temperature relation
    # R_T = R_0 * (1 + A*T + B*T^2 + C*(T-100)*T^3)
    # R_T: resistance at T 'C   [Ohm]
    # R_0: resistance at 0 'C   [Ohm]
    # T  : temperature          ['C]
    # A = 3.9083e-3
    # B = -5.775e-7
    # C = -4.183e-12  # when below 0 'C, C = 0 when above 0 'C

    if R_T >= R_0:
        # We are in the range T >= 0'C
        # Hence, simply solve quadratic equation because C = 0
        sqrt_arg = A ** 2 - 4 * B * (1 - R_T / R_0)
        if sqrt_arg < 0:
            return np.nan
        else:
            T = (-A + np.sqrt(sqrt_arg)) / (2 * B)

    else:
        # We are in the range T < 0'C, hence we need to solve a quartic
        # equation. Difficult to solve by algebra. We do it numerically
        # up to a certain convergence error. A convergence of
        # 0.1 milli-Kelvin is more than sufficient for the PT-104 logger.
        CONV = 1e-4  # [K]
        # Restrict the number of iterations. We have convergence at
        # CONV = 1e-4 within 20 iterations for a PT100 sensor
        MAX_ITER = 40

        # Start iteration loop
        # fmt: off
        T_lo = T_MIN        # Lower bound temperature   ['C]
        T_hi = 0            # Upper bound temperature   ['C]
        T_g  = -1.0         # Initial guess temperature ['C]

        i = 0               # Iteration counter
        diff = 2 * CONV     # = 2 * CONV assures at least 1 iteration
        # fmt: on
        while diff > CONV:
            # Calculate resistance corresponding to the guessed temperature T_g
            R_g = ITS90_degC_to_Ohm(R_0, T_g)

            # How far are we off the reported R_T?
            diff = R_T - R_g

            if (diff > 0) & (abs(diff) > CONV):
                T_lo = T_g
            elif (diff <= 0) & (abs(diff) > CONV):
                T_hi = T_g
                diff = -diff

            # Next best guess
            T_g = (T_hi + T_lo) / 2.0

            i += 1
            if i > MAX_ITER:
                print(
                    "WARNING: Loop in ITS90_Ohm_to_degC() terminated after "
                    "%d iterations" % MAX_ITER
                )
                break

        # We have reached convergence
        T = T_g

    if T > T_MAX:
        print("WARNING: Temperature is out of range because > %.0f 'C" % T_MAX)
    elif T <= T_MIN + 1e-4:
        print("WARNING: Temperature is out of range because <= %.0f 'C" % T_MIN)

    return T


# ------------------------------------------------------------------------------
#   Debug functions
# ------------------------------------------------------------------------------


def print_as_hex(byte_list):
    list(map(lambda x: print(format(x, "02x"), end=" "), byte_list))
    print()


# ------------------------------------------------------------------------------
#   Main
# ------------------------------------------------------------------------------

if __name__ == "__main__":
    import sys

    IP_ADDRESS = "10.10.100.2"
    PORT = 1234

    # Initialise PT104 instance
    pt104 = Picotech_PT104()

    # Connect over UDP port and try to achieve a lock
    if pt104.connect(IP_ADDRESS, PORT):
        pt104.begin()
        pt104.report_EEPROM()
    else:
        print("ERROR: Could not connect to PT-104 and acquire a lock.")
        sys.exit(0)

    # Start the conversion (DAQ)
    if pt104.start_conversion(
        ENA_channels=[1, 1, 0, 0], gain_channels=[1, 1, 0, 0]
    ):
        print("\nConverting")
    else:
        print("\nERROR: Failed start_conversion()")

    # Continuous reading as fast as possible
    print("\nT1 ['C]\tT2 ['C]")
    while 1:
        pt104.scan_4_wire_temperature()
        print("\r%.3f\t%.3f" % (pt104.state.ch1_T, pt104.state.ch2_T), end="")




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""PyQt5 module to provide multithreaded communication and periodical data
acquisition for a Compax3 traverse controller.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "06-07-2020"  # 0.0.1 was stamped 14-09-2018
__version__ = "0.0.5"  # 0.0.1 corresponds to prototype 1.0.0
# pylint: disable=try-except-raise

from PyQt5 import QtCore, QtGui
from PyQt5 import QtWidgets as QtWid

from dvg_utils.dvg_pyqt_controls import (
    SS_GROUP,
    SS_TEXTBOX_ERRORS,
    create_error_LED,
    create_tiny_LED,
)

from dvg_qdeviceio import QDeviceIO, DAQ_TRIGGER
from dvg_devices.Compax3_servo_protocol_RS232 import Compax3_servo


class Compax3_servo_qdev(QDeviceIO):
    """Manages multithreaded communication and periodical data acquisition for
    a Compax3 traverse controller, referred to as the 'device'.

    In addition, it also provides PyQt5 GUI objects for control of the device.
    These can be incorporated into your application.

    All device I/O operations will be offloaded to 'workers', each running in
    a newly created thread instead of in the main/GUI thread.

        - Worker_DAQ:
            Periodically acquires data from the device.

        - Worker_jobs:
            Maintains a thread-safe queue where desired device I/O operations
            can be put onto, and sends the queued operations first in first out
            (FIFO) to the device.

    (*): See 'dvg_qdeviceio.QDeviceIO()' for details.

    Args:
        dev:
            Reference to a 'dvg_devices.Compax3_servo_protocol_RS232.Compax3_servo'
            instance.

        (*) DAQ_interval_ms
        (*) critical_not_alive_count
        (*) DAQ_timer_type

    Main methods:
        (*) start(...)
        (*) quit()

    Main data attributes:
        (*) DAQ_update_counter
        (*) obtained_DAQ_interval_ms
        (*) obtained_DAQ_rate_Hz

    Main GUI objects:
        qgrp (PyQt5.QtWidgets.QGroupBox)

    Signals:
        (*) signal_DAQ_updated()
        (*) signal_connection_lost()
    """

    def __init__(
        self,
        dev: Compax3_servo,
        DAQ_interval_ms=250,
        critical_not_alive_count=1,
        DAQ_timer_type=QtCore.Qt.CoarseTimer,
        calc_DAQ_rate_every_N_iter=4,
        debug=False,
        **kwargs,
    ):
        super().__init__(dev, **kwargs)  # Pass kwargs onto QtCore.QObject()

        self.create_worker_DAQ(
            DAQ_trigger=DAQ_TRIGGER.INTERNAL_TIMER,
            DAQ_function=self.DAQ_function,
            DAQ_interval_ms=DAQ_interval_ms,
            DAQ_timer_type=DAQ_timer_type,
            critical_not_alive_count=critical_not_alive_count,
            calc_DAQ_rate_every_N_iter=calc_DAQ_rate_every_N_iter,
            debug=debug,
        )

        self.create_worker_jobs(debug=debug)

        self.create_GUI()
        self.signal_DAQ_updated.connect(self.update_GUI)
        self.connect_signals_to_slots()
        if not self.dev.is_alive:
            self.update_GUI()  # Correctly reflect an offline device

        # Flags for Jog+/Jog- pushbutton control
        self.jog_plus_is_active = False
        self.jog_minus_is_active = False

    # --------------------------------------------------------------------------
    #   DAQ_function
    # --------------------------------------------------------------------------

    def DAQ_function(self):
        success = self.dev.query_position()
        success &= self.dev.query_status_word_1()

        if not self.dev.status_word_1.no_error:
            self.dev.query_error()

        return success

    # --------------------------------------------------------------------------
    #   create_GUI
    # --------------------------------------------------------------------------

    def create_GUI(self):
        default_font_height = 17
        default_font_width = 8

        # Sub-groupbox: Status word 1 bits
        self.sw1_powerless = create_tiny_LED()
        self.sw1_powered_stationary = create_tiny_LED()
        self.sw1_zero_pos_known = create_tiny_LED()
        self.sw1_pos_reached = create_tiny_LED()

        i = 0
        p = {"alignment": QtCore.Qt.AlignRight}
        grid = QtWid.QGridLayout()
        grid.setVerticalSpacing(4)
        # fmt: off
        grid.addWidget(QtWid.QLabel("powerless", **p)         , i, 0)
        grid.addWidget(self.sw1_powerless                     , i, 1); i+=1
        grid.addWidget(QtWid.QLabel("powered stationary", **p), i, 0)
        grid.addWidget(self.sw1_powered_stationary            , i, 1); i+=1
        grid.addWidget(QtWid.QLabel("zero pos. known", **p)   , i, 0)
        grid.addWidget(self.sw1_zero_pos_known                , i, 1); i+=1
        grid.addWidget(QtWid.QLabel("position reached", **p)  , i, 0)
        grid.addWidget(self.sw1_pos_reached                   , i, 1); i+=1
        # fmt: on
        # grid.setColumnStretch(0, 0)
        # grid.setColumnStretch(1, 0)

        self.qgrp_sw1 = QtWid.QGroupBox("Status word 1")
        self.qgrp_sw1.setLayout(grid)

        # Main groupbox
        font_lbl_status = QtGui.QFont("Palatino", 14, weight=QtGui.QFont.Bold)
        self.lbl_status = QtWid.QLabel(
            "OFFLINE", font=font_lbl_status, alignment=QtCore.Qt.AlignCenter
        )
        self.lbl_status.setFixedHeight(
            3 * QtGui.QFontMetrics(font_lbl_status).height()
        )

        self.sw1_error_tripped = create_error_LED(text="No error")
        self.error_msg = QtWid.QPlainTextEdit("", lineWrapMode=True)
        self.error_msg.setStyleSheet(SS_TEXTBOX_ERRORS)
        self.error_msg.setMinimumWidth(22 * default_font_width)
        self.error_msg.setFixedHeight(4 * default_font_height)
        self.pbtn_ackn_error = QtWid.QPushButton("Acknowledge error")
        self.qled_cur_pos = QtWid.QLineEdit(
            "nan", readOnly=True, alignment=QtCore.Qt.AlignRight
        )
        self.qled_new_pos = QtWid.QLineEdit(
            "nan", readOnly=False, alignment=QtCore.Qt.AlignRight
        )
        self.pbtn_move_to_new_pos = QtWid.QPushButton("Move to new position")
        self.pbtn_move_to_new_pos.setFixedHeight(3 * default_font_height)
        self.pbtn_jog_plus = QtWid.QPushButton("Jog +")
        self.pbtn_jog_minus = QtWid.QPushButton("Jog -")
        self.pbtn_stop = QtWid.QPushButton("\nSTOP &&\nREMOVE POWER\n")
        self.lbl_update_counter = QtWid.QLabel("0")

        i = 0
        p = {"alignment": QtCore.Qt.AlignRight}
        grid = QtWid.QGridLayout()
        grid.setVerticalSpacing(4)
        # fmt: off
        grid.addWidget(self.lbl_status             , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 12)      , i, 0)      ; i+=1
        grid.addWidget(self.sw1_error_tripped      , i, 0, 1, 3); i+=1
        grid.addWidget(self.error_msg              , i, 0, 1, 3); i+=1
        grid.addWidget(self.pbtn_ackn_error        , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 12)      , i, 0)      ; i+=1
        grid.addWidget(self.qgrp_sw1               , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 12)      , i, 0)      ; i+=1

        grid.addWidget(QtWid.QLabel("Position:")   , i, 0, 1, 3); i+=1
        grid.addWidget(QtWid.QLabel("Current")     , i, 0)
        grid.addWidget(self.qled_cur_pos           , i, 1)
        grid.addWidget(QtWid.QLabel("mm")          , i, 2)      ; i+=1
        grid.addWidget(QtWid.QLabel("New")         , i, 0)
        grid.addWidget(self.qled_new_pos           , i, 1)
        grid.addWidget(QtWid.QLabel("mm")          , i, 2)      ; i+=1

        grid.addItem(QtWid.QSpacerItem(1, 12)      , i, 0)      ; i+=1
        grid.addWidget(self.pbtn_move_to_new_pos   , i, 0, 1, 3); i+=1
        grid.addWidget(self.pbtn_jog_plus          , i, 0, 1, 3); i+=1
        grid.addWidget(self.pbtn_jog_minus         , i, 0, 1, 3); i+=1
        grid.addWidget(self.pbtn_stop              , i, 0, 1, 3); i+=1
        grid.addWidget(self.lbl_update_counter     , i, 0, 1, 3); i+=1
        # fmt: on
        # grid.setColumnStretch(0, 0)
        # grid.setColumnStretch(1, 0)
        # grid.setColumnStretch(2, 0)

        self.qgrp = QtWid.QGroupBox("%s" % self.dev.name)
        self.qgrp.setStyleSheet(SS_GROUP)
        self.qgrp.setLayout(grid)
        self.qgrp.setMaximumWidth(200)  # Work=around, hard limit width

    # --------------------------------------------------------------------------
    #   update_GUI
    # --------------------------------------------------------------------------

    @QtCore.pyqtSlot()
    def update_GUI(self):
        """NOTE: 'self.dev.mutex' is not being locked, because we are only
        reading 'state' for displaying purposes. We can do this because 'state'
        members are written and read atomicly.
        Not locking the mutex might speed up the program.
        """
        if self.dev.is_alive:
            # At startup
            if self.DAQ_update_counter == 1:
                self.qled_new_pos.setText("%.2f" % self.dev.state.cur_pos)

            if self.dev.status_word_1.powerless:
                self.lbl_status.setText("powerless")
            else:
                if self.dev.status_word_1.powered_stationary:
                    self.lbl_status.setText("POWERED\nstationary")
                else:
                    self.lbl_status.setText("POWERED")

            self.sw1_error_tripped.setChecked(
                not (self.dev.status_word_1.no_error)
            )
            if self.dev.status_word_1.no_error:
                self.sw1_error_tripped.setText("No error")
                self.error_msg.setPlainText("")
                self.error_msg.setReadOnly(False)
                self.error_msg.setStyleSheet(SS_TEXTBOX_ERRORS)
            else:
                self.sw1_error_tripped.setText("ERROR TRIPPED")
                self.error_msg.setPlainText(self.dev.state.error_msg)
                self.error_msg.setReadOnly(True)
                self.error_msg.setStyleSheet(SS_TEXTBOX_ERRORS)
            self.sw1_powerless.setChecked(self.dev.status_word_1.powerless)
            self.sw1_powered_stationary.setChecked(
                self.dev.status_word_1.powered_stationary
            )
            self.sw1_zero_pos_known.setChecked(
                self.dev.status_word_1.zero_pos_known
            )
            self.sw1_pos_reached.setChecked(self.dev.status_word_1.pos_reached)
            self.qled_cur_pos.setText("%.2f" % self.dev.state.cur_pos)

            self.lbl_update_counter.setText("%s" % self.DAQ_update_counter)
        else:
            self.qgrp.setEnabled(False)

    # --------------------------------------------------------------------------
    #   GUI functions
    # --------------------------------------------------------------------------

    @QtCore.pyqtSlot()
    def process_pbtn_ackn_error(self):
        self.send(self.dev.acknowledge_error)

    @QtCore.pyqtSlot()
    def process_editingFinished_qled_new_pos(self):
        try:
            new_pos = float(self.qled_new_pos.text())
        except (TypeError, ValueError):
            new_pos = 0.0
        except:
            raise
        self.qled_new_pos.setText("%.2f" % new_pos)

    @QtCore.pyqtSlot()
    def process_pbtn_move_to_new_pos(self):
        # Double check if the value in the QLineEdit is actually numeric
        try:
            new_pos = float(self.qled_new_pos.text())
        except:
            raise
        self.send(self.dev.move_to_target_position, (new_pos, 2))

    @QtCore.pyqtSlot()
    def process_pbtn_jog_plus_pressed(self):
        if not self.jog_plus_is_active:
            self.jog_plus_is_active = True
            self.send(self.dev.jog_plus)

    @QtCore.pyqtSlot()
    def process_pbtn_jog_plus_released(self):
        self.jog_plus_is_active = False
        self.send(self.dev.stop_motion_but_keep_power)

    @QtCore.pyqtSlot()
    def process_pbtn_jog_minus_pressed(self):
        if not self.jog_minus_is_active:
            self.jog_minus_is_active = True
            self.send(self.dev.jog_minus)

    @QtCore.pyqtSlot()
    def process_pbtn_jog_minus_released(self):
        self.jog_minus_is_active = False
        self.send(self.dev.stop_motion_but_keep_power)

    @QtCore.pyqtSlot()
    def process_pbtn_stop(self):
        self.send(self.dev.stop_motion_and_remove_power)

    # --------------------------------------------------------------------------
    #   connect_signals_to_slots
    # --------------------------------------------------------------------------

    def connect_signals_to_slots(self):
        # self.send_setpoint.editingFinished.connect(
        #        self.send_setpoint_from_textbox)

        self.pbtn_ackn_error.clicked.connect(self.process_pbtn_ackn_error)
        self.qled_new_pos.editingFinished.connect(
            self.process_editingFinished_qled_new_pos
        )
        self.pbtn_move_to_new_pos.clicked.connect(
            self.process_pbtn_move_to_new_pos
        )
        # self.pbtn_debug.clicked.connect(self.process_pbtn_debug)
        self.pbtn_jog_plus.pressed.connect(self.process_pbtn_jog_plus_pressed)
        self.pbtn_jog_plus.released.connect(self.process_pbtn_jog_plus_released)
        self.pbtn_jog_minus.pressed.connect(self.process_pbtn_jog_minus_pressed)
        self.pbtn_jog_minus.released.connect(
            self.process_pbtn_jog_minus_released
        )
        self.pbtn_stop.clicked.connect(self.process_pbtn_stop)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Class ChartHistory manages two data arrays `x` and `y` that serve as history
buffers for displaying e.g. a strip chart or Lissajous curve. It is
thread-safe and relies on the PyQtGraph library for fast plotting to screen.
Intended multithreaded operation: One thread does the data acquisition and
pushes new data points into the history buffers by calling `add_new_reading(s)`,
and another thread performs the GUI refresh and redraws the data behind the plot
by calling `update_curve`.

New readings are stored in an history array of fixed size. Newest readings
are placed at the end of the array. Array full? -> FIFO.

Class:
    ChartHistory(chart_history_length, plot_data_item):
        Args:
            chart_history_length:
                Number of data points to store in each history buffer.
            plot_data_item:
                Instance of `pyqtgraph.PlotDataItem` to plot out the buffered
                data to.

        Methods:
            apply_downsampling(...):
                Calls the downsampling routines of PyQtGraph, but in the future
                I will provide my own routines for downsampling here.
            add_new_reading(...):
                Add single data point (x, y) to the history buffers.
            add_new_readings(...):
                Add lists of data points (list_x, list_y) to the history
                buffers.
            update_curve():
                Update the data behind the curve and redraw.
            clear():
                Clear buffers.

        Important member:
            x_axis_divisor:
                If the x-data is time, you can use this divisor value to
                transform the x-axis units from e.g. milliseconds to seconds or
                minutes.
            y_axis_divisor:
                Same functionality as x_axis_divisor
"""
__author__      = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__         = "https://github.com/Dennis-van-Gils/DvG_PyQt_misc"
__date__        = "14-09-2018"
__version__     = "1.0.0"

import collections

import numpy as np
from PyQt5 import QtCore
import pyqtgraph as pg

class ChartHistory(object):
    def __init__(self,
                 chart_history_length,
                 plot_data_item: pg.PlotDataItem=None):
        self.chart_history_length = chart_history_length
        self.curve = plot_data_item   # Instance of [pyqtgraph.PlotDataItem]
        self.mutex = QtCore.QMutex()  # For the case of multithreaded access

        # If the x-data is time, you can use this divisor value to transform
        # the x-axis units from e.g. milliseconds to seconds or minutes.
        self.x_axis_divisor = 1
        self.y_axis_divisor = 1

        self._x = collections.deque(maxlen=chart_history_length)
        self._y = collections.deque(maxlen=chart_history_length)
        self._x_snapshot = [0]
        self._y_snapshot = [0]

        # Performance boost: Do not plot data outside of visible range
        self.curve.clipToView = True

        # Default to no downsampling
        self.curve.setDownsampling(ds=1, auto=False, method='mean')

    def apply_downsampling(self, do_apply=True, ds=4):
        if do_apply:
            # Speed up plotting, needed for keeping the GUI responsive when
            # using large datasets
            self.curve.setDownsampling(ds=ds, auto=False, method='mean')
        else:
            self.curve.setDownsampling(ds=1, auto=False, method='mean')

    def add_new_reading(self, x, y):
        locker = QtCore.QMutexLocker(self.mutex)
        self._x.append(x)
        self._y.append(y)
        locker.unlock()

    def add_new_readings(self, x_list, y_list):
        locker = QtCore.QMutexLocker(self.mutex)
        self._x.extend(x_list)
        self._y.extend(y_list)
        locker.unlock()

    def update_curve(self):
        """Creates a snapshot of the buffered data, which is a fast operation,
        followed by updating the data behind the curve and redrawing it, which
        is a slow operation. Hence, the use of a snapshot creation, which is
        locked my a mutex, followed by a the mutex unlocked redrawing.
        """

        # First create a snapshot of the buffered data. Fast.
        locker = QtCore.QMutexLocker(self.mutex)
        self._x_snapshot = np.copy(self._x)
        self._y_snapshot = np.copy(self._y)
        #print("numel x: %d, numel y: %d" %
        #      (self._x_snapshot.size, self._y_snapshot.size))
        locker.unlock()

        # Now update the data behind the curve and redraw the curve. Slow
        if self.curve is not None:
            if ((len(self._x_snapshot) == 0) or
                (np.alltrue(np.isnan(self._y_snapshot)))):
                self.curve.setData([0], [0])
            else:
                self.curve.setData((self._x_snapshot - self._x_snapshot[-1])
                                   / float(self.x_axis_divisor),
                                   self._y_snapshot
                                   / float(self.y_axis_divisor))

    def clear(self):
        locker = QtCore.QMutexLocker(self.mutex)
        self._x.clear()
        self._y.clear()
        locker.unlock()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Multithreaded PyQt5 GUI to interface with a Compax3 traverse controller.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "06-07-2020"  # 0.0.1 was stamped 14-09-2018
__version__ = "0.0.5"  # 0.0.1 corresponds to prototype 1.0.0
# pylint: disable=bare-except

import sys
from pathlib import Path

from PyQt5 import QtCore, QtGui
from PyQt5 import QtWidgets as QtWid

from dvg_utils.dvg_pyqt_controls import SS_TEXTBOX_READ_ONLY, SS_GROUP

from dvg_devices.Compax3_servo_protocol_RS232 import Compax3_servo
from dvg_devices.Compax3_servo_qdev import Compax3_servo_qdev
from dvg_devices.Compax3_servo_step_navigator_GUI import Compax3_step_navigator

# ------------------------------------------------------------------------------
#   MainWindow
# ------------------------------------------------------------------------------


class MainWindow(QtWid.QWidget):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        self.setGeometry(40, 60, 0, 0)
        self.setWindowTitle("Compax3 traverse controller")

        # Top grid
        self.lbl_title = QtWid.QLabel(
            "Compax3 traverse controller",
            font=QtGui.QFont("Palatino", 14, weight=QtGui.QFont.Bold),
        )
        self.pbtn_exit = QtWid.QPushButton("Exit")
        self.pbtn_exit.clicked.connect(self.close)
        self.pbtn_exit.setMinimumHeight(30)

        grid_top = QtWid.QGridLayout()
        grid_top.addWidget(self.lbl_title, 0, 0)
        grid_top.addWidget(self.pbtn_exit, 0, 1, QtCore.Qt.AlignRight)

        # Traverse schematic image
        lbl_trav_img = QtWid.QLabel()
        lbl_trav_img.setPixmap(
            QtGui.QPixmap(
                str(Path(sys.modules[__name__].__file__).parent)
                + "/Traverse_layout.png"
            )
        )
        lbl_trav_img.setFixedSize(244, 240)

        grid = QtWid.QGridLayout()
        grid.addWidget(lbl_trav_img, 0, 0, QtCore.Qt.AlignTop)

        grpb_trav_img = QtWid.QGroupBox("Traverse schematic")
        grpb_trav_img.setStyleSheet(SS_GROUP)
        grpb_trav_img.setLayout(grid)

        # Round up full window
        vbox = QtWid.QVBoxLayout()
        vbox.addWidget(grpb_trav_img)
        vbox.addWidget(trav_step_nav.grpb)
        vbox.addStretch(1)
        vbox.setAlignment(trav_step_nav.grpb, QtCore.Qt.AlignLeft)

        hbox = QtWid.QHBoxLayout()
        hbox.addWidget(trav_vert_qdev.qgrp)
        hbox.addWidget(trav_horz_qdev.qgrp)
        hbox.addLayout(vbox)
        hbox.addStretch(1)
        hbox.setAlignment(trav_horz_qdev.qgrp, QtCore.Qt.AlignTop)
        hbox.setAlignment(trav_vert_qdev.qgrp, QtCore.Qt.AlignTop)

        vbox = QtWid.QVBoxLayout(self)
        vbox.addLayout(grid_top)
        vbox.addLayout(hbox)


# ------------------------------------------------------------------------------
#   Act on step signals
# ------------------------------------------------------------------------------


@QtCore.pyqtSlot()
def act_upon_signal_step_up(new_pos: float):
    trav_vert_qdev.qled_new_pos.setText("%.2f" % new_pos)
    trav_vert_qdev.process_pbtn_move_to_new_pos()


@QtCore.pyqtSlot()
def act_upon_signal_step_down(new_pos: float):
    trav_vert_qdev.qled_new_pos.setText("%.2f" % new_pos)
    trav_vert_qdev.process_pbtn_move_to_new_pos()


@QtCore.pyqtSlot()
def act_upon_signal_step_left(new_pos: float):
    trav_horz_qdev.qled_new_pos.setText("%.2f" % new_pos)
    trav_horz_qdev.process_pbtn_move_to_new_pos()


@QtCore.pyqtSlot()
def act_upon_signal_step_right(new_pos: float):
    trav_horz_qdev.qled_new_pos.setText("%.2f" % new_pos)
    trav_horz_qdev.process_pbtn_move_to_new_pos()


# ------------------------------------------------------------------------------
#   about_to_quit
# ------------------------------------------------------------------------------


def about_to_quit():
    print("About to quit")
    app.processEvents()

    for trav_qdev in travs_qdev:
        trav_qdev.quit()

    for trav in travs:
        trav.close()


# ------------------------------------------------------------------------------
#   Main
# ------------------------------------------------------------------------------

if __name__ == "__main__":

    # Specific connection settings of each traverse axis of our setup
    class Trav_connection_params:
        # Serial number of the Compax3 traverse controller to connect to.
        # Set to '' or None to connect to any Compax3.
        serial = None
        # Display name
        name = "TRAV"
        # Path to the config textfile containing the (last used) RS232 port
        path_config = Path("config/port_Compax3_trav.txt")

    # Horizontal axis
    trav_conn_horz = Trav_connection_params()
    trav_conn_horz.serial = "4409980001"
    trav_conn_horz.name = "TRAV HORZ"
    trav_conn_horz.path_config = Path("config/port_Compax3_trav_horz.txt")

    # Vertical axis
    trav_conn_vert = Trav_connection_params()
    trav_conn_vert.serial = "4319370001"
    trav_conn_vert.name = "TRAV VERT"
    trav_conn_vert.path_config = Path("config/port_Compax3_trav_vert.txt")

    # The state of the traverse controllers is polled with this time interval
    UPDATE_INTERVAL_MS = 250  # [ms]

    # --------------------------------------------------------------------------
    #   Connect to and set up Compax3 traverse controllers
    # --------------------------------------------------------------------------

    trav_horz = Compax3_servo(name=trav_conn_horz.name)
    trav_vert = Compax3_servo(name=trav_conn_vert.name)

    if trav_horz.auto_connect(
        trav_conn_horz.path_config, trav_conn_horz.serial
    ):
        trav_horz.begin()

        # Set the default motion profile (= #2) parameters
        trav_horz.store_motion_profile(
            target_position=0,
            velocity=10,
            mode=1,
            accel=100,
            decel=100,
            jerk=1e6,
            profile_number=2,
        )

    if trav_vert.auto_connect(
        trav_conn_vert.path_config, trav_conn_vert.serial
    ):
        trav_vert.begin()

        # Set the default motion profile (= #2) parameters
        trav_vert.store_motion_profile(
            target_position=0,
            velocity=10,
            mode=1,
            accel=100,
            decel=100,
            jerk=1e6,
            profile_number=2,
        )

    travs = [trav_horz, trav_vert]

    # --------------------------------------------------------------------------
    #   Create application
    # --------------------------------------------------------------------------
    QtCore.QThread.currentThread().setObjectName("MAIN")  # For DEBUG info

    app = 0  # Work-around for kernel crash when using Spyder IDE
    app = QtWid.QApplication(sys.argv)
    app.setFont(QtGui.QFont("Arial", 9))
    app.setStyleSheet(SS_TEXTBOX_READ_ONLY)
    app.aboutToQuit.connect(about_to_quit)

    # Create PyQt GUI interfaces and communication threads for the device
    trav_horz_qdev = Compax3_servo_qdev(trav_horz, UPDATE_INTERVAL_MS)
    trav_vert_qdev = Compax3_servo_qdev(trav_vert, UPDATE_INTERVAL_MS)
    travs_qdev = [trav_horz_qdev, trav_vert_qdev]

    # Create Compax3 single step navigator
    trav_step_nav = Compax3_step_navigator(
        trav_horz=trav_horz, trav_vert=trav_vert
    )
    trav_step_nav.step_up.connect(act_upon_signal_step_up)
    trav_step_nav.step_down.connect(act_upon_signal_step_down)
    trav_step_nav.step_left.connect(act_upon_signal_step_left)
    trav_step_nav.step_right.connect(act_upon_signal_step_right)

    # Create window
    window = MainWindow()

    # --------------------------------------------------------------------------
    #   Start threads
    # --------------------------------------------------------------------------

    trav_horz_qdev.start()
    trav_vert_qdev.start()

    # --------------------------------------------------------------------------
    #   Start the main GUI event loop
    # --------------------------------------------------------------------------

    window.show()
    sys.exit(app.exec_())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Class FileLogger handles logging data to file, particularly well suited for
multithreaded programs, where one thread is writing data to the log (the logging
thread) and the other thread (the main thread/GUI) handles starting and stopping
of the logging by user interaction (i.e. a button).
    
The boolean members 'starting' and 'stopping' should be directly written to from
the main/GUI thread. (These are boolean atomic operations so no mutex lock is
necessary here.)

In the logging thread one should test for these booleans as demonstrated in the
following example:
    if file_logger.starting:
        if file_logger.create_log(my_current_time, my_path):
            file_logger.write("Time\tValue\n")  # Header
        
    if file_logger.stopping:
        file_logger.close_log()
        
    if file_logger.is_recording:
        elapsed_time = my_current_time - file_logger.start_time
        file_logger.write("%.3f\t%.3f\n" % (elapsed_time, my_value))

Class:
    FileLogger():   
        Methods:
            create_log(...):
                Open new log file and keep file handle open.
            write(...):
                Write data to the open log file.
            close_log():
                Close the log file.
                
        Important members:
            starting (bool):
            stopping (bool):
            is_recording (bool):

        Signals:
            signal_set_recording_text(str):
                Useful for updating text of e.g. a record button when using a
                PyQt GUI. This signal is not emitted in this module itself, and
                you should emit it yourself in your own code when needed.
"""
__author__      = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__         = "https://github.com/Dennis-van-Gils/DvG_PyQt_misc"
__date__        = "23-08-2018"
__version__     = "1.0.0"

from pathlib import Path

from PyQt5 import QtCore

from dvg_debug_functions import print_fancy_traceback as pft

class FileLogger(QtCore.QObject):
    signal_set_recording_text = QtCore.pyqtSignal(str)

    def __init__(self):
        super().__init__(None)

        self.path_log = None        # pathlib.Path instance to the log
        self.f_log = None           # File handle to log
        
        self.start_time = None      # To keep track of elapsed time since start
        self.starting = False
        self.stopping = False
        self.is_recording = False
        
        # Placeholder for a future mutex instance needed for proper
        # multithreading (e.g. instance of QtCore.Qmutex())
        self.mutex = None
        
    def create_log(self, start_time, path_log: Path, mode='a'):
        """Open new log file and keep file handle open.
        
        Args:
            start_time:
                Timestamp of the start of recording, usefull to keep track of
                the elapsed time while recording.
            path_log [pathlib.Path]:
                Location of the file to write to.
            mode:
                Mode in which the file is openend, see 'open()' for more
                details. Defaults to 'a'. Most common options:
                'w': Open for writing, truncating the file first
                'a': Open for writing, appending to the end of the file if it
                     exists
        
        Returns: True if successful, False otherwise.
        """
        self.path_log = path_log
        self.start_time = start_time
        self.starting = False
        self.stopping = False
        
        try:
            self.f_log = open(path_log, mode)
        except Exception as err:
            pft(err, 3)
            self.is_recording = False
            return False
        else:            
            self.is_recording = True
            return True
    
    def write(self, data):
        """
        Returns: True if successful, False otherwise.
        """
        try:
            self.f_log.write(data)
        except Exception as err:
            pft(err, 3)
            return False
        else:
            return True
        
    def close_log(self):
        if self.is_recording:
            self.f_log.close()
        self.starting = False
        self.stopping = False
        self.is_recording = False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""PyQt5 module to provide multithreaded communication and periodical data
acquisition for a Keysight (former HP or Agilent) 34970A/34972A data
acquisition/switch unit. Different boards can be installed inside such a unit.
This library is intended to be used with multiplexer board(s), as it will
scan over the board's input channels to retrieve the readings. Hence, we also
refer to this device as a multiplexer, or mux.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "07-07-2020"  # 0.0.1 was stamped 18-09-2018
__version__ = "0.0.5"  # 0.0.1 corresponds to prototype 1.0.0


from PyQt5 import QtCore, QtGui
from PyQt5 import QtWidgets as QtWid

from dvg_utils.dvg_pyqt_controls import (
    create_Toggle_button,
    SS_TEXTBOX_ERRORS,
    SS_TEXTBOX_READ_ONLY,
    SS_GROUP,
)
from dvg_debug_functions import dprint, print_fancy_traceback as pft

from dvg_qdeviceio import QDeviceIO, DAQ_TRIGGER
from dvg_devices.Keysight_3497xA_protocol_SCPI import Keysight_3497xA


# Monospace font
FONT_MONOSPACE = QtGui.QFont("Monospace", 12, weight=QtGui.QFont.Bold)
FONT_MONOSPACE.setStyleHint(QtGui.QFont.TypeWriter)

FONT_MONOSPACE_SMALL = QtGui.QFont("Monospace", 9)
FONT_MONOSPACE_SMALL.setStyleHint(QtGui.QFont.TypeWriter)

# Infinity cap: values reported by the 3497xA greater than this will be
# displayed as 'inf'
INFINITY_CAP = 9.8e37

# Short-hand alias for DEBUG information
def get_tick():
    return QtCore.QDateTime.currentMSecsSinceEpoch()


class Keysight_3497xA_qdev(QDeviceIO):
    """Manages multithreaded communication and periodical data acquisition for
    a Keysight (former HP or Agilent) 34970A/34972A data acquisition/switch
    unit, referred to as the 'device'. Different boards can be installed inside
    such a unit. This library is intended to be used with multiplexer board(s),
    as it will scan over the board's input channels to retrieve the readings.
    Hence, we also refer to this device as a multiplexer, or mux.

    In addition, it also provides PyQt5 GUI objects for control of the device.
    These can be incorporated into your application.

    All device I/O operations will be offloaded to 'workers', each running in
    a newly created thread instead of in the main/GUI thread.

        - Worker_DAQ:
            Periodically acquires data from the device.

        - Worker_jobs:
            Maintains a thread-safe queue where desired device I/O operations
            can be put onto, and sends the queued operations first in first out
            (FIFO) to the device.

    (*): See 'dvg_qdeviceio.QDeviceIO()' for details.

    Args:
        dev:
            Reference to a
            'dvg_devices.Keysight_3497xA_protocol_SCPI.Keysight_3497xA' instance.

        (*) DAQ_interval_ms
        (*) critical_not_alive_count
        (*) DAQ_timer_type

        DAQ_postprocess_MUX_data_function (optional, default=None):
            Reference to a user-supplied function that will be called during
            an 'worker_DAQ' update, after a mux scan has been performed. Hence,
            you can use this function to, e.g., parse out the scan readings into
            separate variables and post-process this data or log it.

    Main methods:
        (*) start(...)
        (*) quit()

        set_table_readings_format:
            TO DO: write description

    Main data attributes:
        is_MUX_scanning (read-only bool):
            True when the multiplexer is fetching new scan data every DAQ
            update, False otherwise. Do not set this variable directly. Use
            the slots 'start_MUX_scan' and 'stop_MUX_scan', instead.

    Main GUI objects:
        qgrp (PyQt5.QtWidgets.QGroupBox)

    Signals:
        (*) signal_DAQ_updated()
        (*) signal_connection_lost()

    Slots:
        start_MUX_scan():
            Enable scanning and fetching data of the multiplexer during an
            'worker_DAQ' update.

        stop_MUX_scan():
            Disable scanning and fetching data of the multiplexer during an
            'worker_DAQ' update.
    """

    def __init__(
        self,
        dev: Keysight_3497xA,
        DAQ_interval_ms=1000,
        DAQ_timer_type=QtCore.Qt.CoarseTimer,
        critical_not_alive_count=3,
        calc_DAQ_rate_every_N_iter=1,
        DAQ_postprocess_MUX_scan_function=None,
        debug=False,
        **kwargs,
    ):
        super().__init__(dev, **kwargs)  # Pass kwargs onto QtCore.QObject()

        self.create_worker_DAQ(
            DAQ_trigger=DAQ_TRIGGER.INTERNAL_TIMER,
            DAQ_function=self.DAQ_function,
            DAQ_interval_ms=DAQ_interval_ms,
            DAQ_timer_type=DAQ_timer_type,
            critical_not_alive_count=critical_not_alive_count,
            calc_DAQ_rate_every_N_iter=calc_DAQ_rate_every_N_iter,
            debug=debug,
        )

        self.create_worker_jobs(jobs_function=self.jobs_function, debug=debug)

        self.DAQ_postprocess_MUX_scan_function = (
            DAQ_postprocess_MUX_scan_function
        )
        self.is_MUX_scanning = False

        # String format to use for the readings in the table widget.
        # When type is a single string, all rows will use this format.
        # When type is a list of strings, rows will be formatted consecutively.
        self.table_readings_format = "%.3e"

        self.create_GUI()
        self.signal_DAQ_updated.connect(self.update_GUI)

        # Populate the table view with QTableWidgetItems.
        # I.e. add the correct number of rows to the table depending on the
        # full scan list.
        self.populate_table_readings()

        # Populate the textbox with the SCPI setup commands
        self.populate_SCPI_commands()

        # Update GUI immediately, instead of waiting for the first refresh
        self.update_GUI()

    # --------------------------------------------------------------------------
    #   start_scanning
    # --------------------------------------------------------------------------

    @QtCore.pyqtSlot()
    def start_MUX_scan(self):
        """
        # [DISABLED CODE]
        # Recreate the worker timer and acquire new samples immediately,
        # instead of waiting for the next tick to occur of the 'old' timer.
        # NOTE: Would love to stop the old timer first, but for some strange
        # reason the very first Qtimer appears to be running in another
        # thread than the 'worker_state' thread, eventhough the routine in
        # K3497xA_qdev.__init__ has finished moving the worker_state thread.
        # When timer is running in another thread than 'this' one, an
        # exception is thrown py Python saying we can't stop the timer from
        # another thread. Hence, I reassign a new Qtimer and hope that the
        # old timer is garbage collected all right.
        self.timer.stop()  # Does not work 100% of the time!
        """

        self.is_MUX_scanning = True
        self.signal_DAQ_updated.emit()  # Show we are scanning
        QtWid.QApplication.processEvents()

        """
        # [DISABLED CODE]
        # Disabled the recreation of the timer, because of not
        # understood behavior. The first time iteration of every newly
        # created QTimer seems to be running in the MAIN thread while
        # subsequent iterations take place in the Worker_state thread.
        # This might lead to program instability? Not sure. Simply disabled
        # for now.
        self.timer = QtCore.QTimer()            # Create new timer
        self.timer.setInterval(self.scanning_interval_ms)
        self.timer.timeout.connect(self.update)
        self.timer.start()
        self.update()  # Kickstart at t = 0, because timer doesn't fire now
        """

    # --------------------------------------------------------------------------
    #   stop_scanning
    # --------------------------------------------------------------------------

    @QtCore.pyqtSlot()
    def stop_MUX_scan(self):
        self.is_MUX_scanning = False
        self.signal_DAQ_updated.emit()  # Show we stopped scanning
        QtWid.QApplication.processEvents()

    # --------------------------------------------------------------------------
    #   DAQ_function
    # --------------------------------------------------------------------------

    def DAQ_function(self):
        tick = get_tick()

        # Clear input and output buffers of the device. Seems to resolve
        # intermittent communication time-outs.
        self.dev.device.clear()

        success = True
        if self.is_MUX_scanning:
            success &= self.dev.init_scan()  # Init scan

            if success:
                self.dev.wait_for_OPC()  # Wait for OPC
                if self.worker_DAQ.debug:
                    tock = get_tick()
                    dprint("opc? in: %i" % (tock - tick))
                    tick = tock

                success &= self.dev.fetch_scan()  # Fetch scan
                if self.worker_DAQ.debug:
                    tock = get_tick()
                    dprint("fetc in: %i" % (tock - tick))
                    tick = tock

            if success:
                self.dev.wait_for_OPC()  # Wait for OPC
                if self.worker_DAQ.debug:
                    tock = get_tick()
                    dprint("opc? in: %i" % (tock - tick))
                    tick = tock

        if success:
            # Do not throw additional timeout exceptions when .init_scan()
            # might have already failed. Hence this check for no success.
            self.dev.query_all_errors_in_queue()  # Query errors
            if self.worker_DAQ.debug:
                tock = get_tick()
                dprint("err? in: %i" % (tock - tick))
                tick = tock

            # The next statement seems to trigger timeout, but very
            # intermittently (~once per 20 minutes). After this timeout,
            # everything times out.
            # self.dev.wait_for_OPC()
            # if self.worker_DAQ.debug:
            #    tock = get_tick()
            #    dprint("opc? in: %i" % (tock - tick))
            #    tick = tock

            # NOTE: Another work-around to intermittent time-outs might
            # be sending self.dev.clear() every iter to clear the input and
            # output buffers. This is now done at the start of this function.

        # Optional user-supplied function to run. You can use this function to,
        # e.g., parse out the scan readings into separate variables and
        # post-process this data or log it.
        if not (self.DAQ_postprocess_MUX_scan_function is None):
            self.DAQ_postprocess_MUX_scan_function()

        if self.worker_DAQ.debug:
            tock = get_tick()
            dprint("extf in: %i" % (tock - tick))
            tick = tock

        return success

    # --------------------------------------------------------------------------
    #   jobs_function
    # --------------------------------------------------------------------------

    def jobs_function(self, func, args):
        # Send I/O operation to the device
        try:
            func(*args)
            self.dev.wait_for_OPC()  # Wait for OPC
        except Exception as err:
            pft(err)

    # --------------------------------------------------------------------------
    #   create_GUI
    # --------------------------------------------------------------------------

    def create_GUI(self):
        p = {"alignment": QtCore.Qt.AlignCenter, "font": FONT_MONOSPACE}
        p2 = {"alignment": QtCore.Qt.AlignCenter + QtCore.Qt.AlignVCenter}
        # self.qlbl_mux = QtWid.QLabel("Keysight 34972a", **p2)
        self.qlbl_mux_state = QtWid.QLabel("Offline", **p)
        self.qpbt_start_scan = create_Toggle_button("Start scan")

        self.qpte_SCPI_commands = QtWid.QPlainTextEdit(
            "", readOnly=True, lineWrapMode=False
        )
        self.qpte_SCPI_commands.setStyleSheet(SS_TEXTBOX_READ_ONLY)
        self.qpte_SCPI_commands.setMaximumHeight(152)
        self.qpte_SCPI_commands.setMinimumWidth(200)
        self.qpte_SCPI_commands.setFont(FONT_MONOSPACE_SMALL)

        p = {"alignment": QtCore.Qt.AlignRight, "readOnly": True}
        self.qled_scanning_interval_ms = QtWid.QLineEdit("", **p)
        self.qled_obtained_interval_ms = QtWid.QLineEdit("", **p)

        self.qpte_errors = QtWid.QPlainTextEdit("", lineWrapMode=False)
        self.qpte_errors.setStyleSheet(SS_TEXTBOX_ERRORS)
        self.qpte_errors.setMaximumHeight(90)

        # fmt: off
        self.qpbt_ackn_errors    = QtWid.QPushButton("Acknowledge errors")
        self.qpbt_reinit         = QtWid.QPushButton("Reinitialize")
        self.qlbl_update_counter = QtWid.QLabel("0")
        self.qpbt_debug_test     = QtWid.QPushButton("Debug test")
        # fmt: on

        self.qpbt_start_scan.clicked.connect(self.process_qpbt_start_scan)
        self.qpbt_ackn_errors.clicked.connect(self.process_qpbt_ackn_errors)
        self.qpbt_reinit.clicked.connect(self.process_qpbt_reinit)
        self.qpbt_debug_test.clicked.connect(self.process_qpbt_debug_test)

        i = 0
        p = {"alignment": QtCore.Qt.AlignLeft + QtCore.Qt.AlignVCenter}

        grid = QtWid.QGridLayout()
        grid.setVerticalSpacing(4)
        # fmt: off
        # grid.addWidget(self.qlbl_mux                    , i, 0, 1, 3); i+=1
        grid.addWidget(QtWid.QLabel("Only scan when necessary.", **p2)
                                                          , i, 0, 1, 2); i+=1
        grid.addWidget(QtWid.QLabel("It wears down the multiplexer.", **p2)
                                                          , i, 0, 1, 2); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 3)              , i, 0)      ; i+=1
        grid.addWidget(self.qlbl_mux_state                , i, 0, 1, 2); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 3)              , i, 0)      ; i+=1
        grid.addWidget(self.qpbt_start_scan               , i, 0, 1, 2); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 4)              , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("SCPI scan commands:"), i, 0, 1, 2); i+=1
        grid.addWidget(self.qpte_SCPI_commands            , i, 0, 1, 2); i+=1
        grid.addWidget(QtWid.QLabel("Scanning interval [ms]"), i, 0)
        grid.addWidget(self.qled_scanning_interval_ms     , i, 1)      ; i+=1
        grid.addWidget(QtWid.QLabel("Obtained [ms]")      , i, 0)
        grid.addWidget(self.qled_obtained_interval_ms     , i, 1)      ; i+=1
        grid.addWidget(self.qpbt_reinit                   , i, 0, 1, 2); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 12)             , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Errors:")            , i, 0, 1, 2); i+=1
        grid.addWidget(self.qpte_errors                   , i, 0, 1, 2); i+=1
        grid.addWidget(self.qpbt_ackn_errors              , i, 0, 1, 2); i+=1
        grid.addWidget(self.qlbl_update_counter           , i, 0, 1, 2); i+=1
        # grid.addWidget(self.qpbt_debug_test             , i, 0, 1, 2); i+=1
        # fmt: on

        #  Table widget containing the readings of the current scan cycle
        # ----------------------------------------------------------------
        self.qtbl_readings = QtWid.QTableWidget(columnCount=1)
        self.qtbl_readings.setHorizontalHeaderLabels(["Readings"])
        self.qtbl_readings.horizontalHeaderItem(0).setFont(FONT_MONOSPACE_SMALL)
        self.qtbl_readings.verticalHeader().setFont(FONT_MONOSPACE_SMALL)
        self.qtbl_readings.verticalHeader().setDefaultSectionSize(24)
        self.qtbl_readings.setFont(FONT_MONOSPACE_SMALL)
        # self.qtbl_readings.setMinimumHeight(600)
        self.qtbl_readings.setFixedWidth(180)
        self.qtbl_readings.setColumnWidth(0, 100)

        grid.addWidget(self.qtbl_readings, 0, 2, i, 1)

        self.qgrp = QtWid.QGroupBox("%s" % self.dev.name)
        self.qgrp.setStyleSheet(SS_GROUP)
        self.qgrp.setLayout(grid)

    # --------------------------------------------------------------------------
    #   update_GUI
    # --------------------------------------------------------------------------

    @QtCore.pyqtSlot()
    def update_GUI(self):
        """NOTE: 'self.dev.mutex' is not being locked, because we are only
        reading 'state' for displaying purposes. We can do this because 'state'
        members are written and read atomicly, with the only exception being
        'str_all_errors', and it bears no consequences to read wrongly.
        Not locking the mutex might speed up the program.
        """
        if self.dev.is_alive:
            if self.is_MUX_scanning:
                self.qlbl_mux_state.setText("Scanning")
                self.qpbt_start_scan.setChecked(True)
            else:
                self.qlbl_mux_state.setText("Idle")
                self.qpbt_start_scan.setChecked(False)

            self.qpte_errors.setReadOnly(self.dev.state.all_errors != [])
            self.qpte_errors.setStyleSheet(SS_TEXTBOX_ERRORS)
            self.qpte_errors.setPlainText(
                "%s" % "\n".join(self.dev.state.all_errors)
            )

            self.qled_obtained_interval_ms.setText(
                "%.0f" % self.obtained_DAQ_interval_ms
            )
            self.qlbl_update_counter.setText("%s" % self.DAQ_update_counter)

            for i in range(len(self.dev.state.all_scan_list_channels)):
                if i >= len(self.dev.state.readings):
                    break
                reading = self.dev.state.readings[i]
                if reading > INFINITY_CAP:
                    self.qtbl_readings.item(i, 0).setData(
                        QtCore.Qt.DisplayRole, "Inf"
                    )
                else:
                    if type(self.table_readings_format) == list:
                        try:
                            str_format = self.table_readings_format[i]
                        except IndexError:
                            str_format = self.table_readings_format[0]
                    elif type(self.table_readings_format) == str:
                        str_format = self.table_readings_format

                    self.qtbl_readings.item(i, 0).setData(
                        QtCore.Qt.DisplayRole, str_format % reading
                    )
        else:
            self.qlbl_mux_state.setText("Offline")
            self.qgrp.setEnabled(False)

    # --------------------------------------------------------------------------
    #   populate_SCPI_commands
    # --------------------------------------------------------------------------

    def populate_SCPI_commands(self):
        self.qpte_SCPI_commands.setPlainText(
            "%s" % "\n".join(self.dev.SCPI_setup_commands)
        )
        self.qled_scanning_interval_ms.setText(
            "%i" % self.worker_DAQ._DAQ_interval_ms
        )

    # --------------------------------------------------------------------------
    #   Table widget related
    # --------------------------------------------------------------------------

    def populate_table_readings(self):
        self.qtbl_readings.setRowCount(
            len(self.dev.state.all_scan_list_channels)
        )
        self.qtbl_readings.setVerticalHeaderLabels(
            self.dev.state.all_scan_list_channels
        )

        for i in range(len(self.dev.state.all_scan_list_channels)):
            item = QtWid.QTableWidgetItem("nan")
            item.setTextAlignment(QtCore.Qt.AlignRight + QtCore.Qt.AlignCenter)
            self.qtbl_readings.setItem(i, 0, item)

    def set_table_readings_format(self, format_str):
        # String format to use for the readings in the table widget
        # When type is a single string, all rows will use this format.
        # When type is a list of strings, rows will be formatted consecutively.
        self.table_readings_format = format_str

    # --------------------------------------------------------------------------
    #   GUI functions
    # --------------------------------------------------------------------------

    @QtCore.pyqtSlot()
    def process_qpbt_start_scan(self):
        if self.qpbt_start_scan.isChecked():
            self.start_MUX_scan()
        else:
            self.stop_MUX_scan()

    @QtCore.pyqtSlot()
    def process_qpbt_ackn_errors(self):
        # Lock the dev mutex because string operations are not atomic
        locker = QtCore.QMutexLocker(self.dev.mutex)
        self.dev.state.all_errors = []
        self.qpte_errors.setPlainText("")
        self.qpte_errors.setReadOnly(False)  # To change back to regular colors
        locker.unlock()

    @QtCore.pyqtSlot()
    def process_qpbt_reinit(self):
        str_msg = (
            "Are you sure you want reinitialize the multiplexer?\n\n"
            "This would abort the current scan, reset the device\n"
            "and resend the SCPI scan command list."
        )
        reply = QtWid.QMessageBox.question(
            None,
            ("Reinitialize %s" % self.dev.name),
            str_msg,
            QtWid.QMessageBox.Yes | QtWid.QMessageBox.No,
            QtWid.QMessageBox.No,
        )

        if reply == QtWid.QMessageBox.Yes:
            self.qpbt_start_scan.setChecked(False)
            self.stop_MUX_scan()
            self.add_to_send_queue(self.dev.wait_for_OPC)
            self.add_to_send_queue(self.dev.begin)
            self.process_send_queue()

    @QtCore.pyqtSlot()
    def process_qpbt_debug_test(self):
        self.send(self.dev.write, "junk")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_namespace_packages
from setuptools import setup


def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")
    ) as fh:
        return fh.read()


setup(
    name="dvg-devices",
    version="0.0.5",
    license="MIT",
    description="Collection of interfaces to communicate with laboratory devices, with support for multithreading in PyQt5.",
    long_description="%s\n%s"
    % (
        re.compile("^.. start-badges.*^.. end-badges", re.M | re.S).sub(
            "", read("README.rst")
        ),
        re.sub(":[a-z]+:`~?(.*?)`", r"``\1``", read("CHANGELOG.rst")),
    ),
    long_description_content_type="text/x-rst",
    author="Dennis van Gils",
    author_email="vangils.dennis@gmail.com",
    url="https://github.com/Dennis-van-Gils/python-dvg-devices",
    packages=find_namespace_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        #'Development Status :: 5 - Production/Stable',
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering ",
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    project_urls={
        "Issue Tracker": "https://github.com/Dennis-van-Gils/python-dvg-devices/issues",
    },
    keywords=[
        "device I/O",
        "automation",
        "laboratory",
        "science",
        "control",
        "experiment",
        "PyQt5",
        "multithread",
    ],
    python_requires=">=3.6",
    install_requires=[
        "dvg-debug-functions~=2.0",
        "dvg-qdeviceio~=0.2.0",
        "pyserial==3.4",
        "pyqt5~=5.12",
        "numpy~=1.15",
        "pyvisa~=1.9",
        "matplotlib~=3.1",
        # "dvg_pyqt_controls",
        # "dvg_pyqt_charthistory",
        # "dvg_pyqt_filelogger",
        
    ],
    extras_require={},
)
